use std::collections::HashMap;
use std::rc::Rc;

use libretrogd::graphics::BitmapAtlas;
use libretrogd::math::*;
use libretrogd::utils::rnd_value;

use crate::logic::*;
use crate::spawning::{AnimalSprites, PeopleSprites};

pub struct World {
    pub person_spawn_time: f32,
    pub animal_spawn_time: f32,
    pub remaining_time: Option<f32>,
    pub time_to_game_over: Option<f32>,
}

impl World {
    pub fn new(time: Option<f32>) -> Self {
        World {
            person_spawn_time: PERSON_SPAWN_TIMER_MAX,
            animal_spawn_time: ANIMAL_SPAWN_TIMER_MIN,
            remaining_time: time,
            time_to_game_over: None,
        }
    }
}

pub struct Player;

pub struct Animal;

pub struct Person;

pub struct Moon {
    pub anger: f32,
    pub time_until_tick: f32,
}

impl Moon {
    pub fn new() -> Self {
        Moon {
            anger: 0.0,
            time_until_tick: MOON_MIN_TIME_BETWEEN_TICKS,
        }
    }

    pub fn update(&mut self, delta: f32) -> bool {
        if self.anger > MOON_MAX_ANGER {
            self.anger = MOON_MAX_ANGER;
        }
        self.anger -= delta;
        if self.anger < 0.0 {
            self.anger = 0.0;
        }

        if self.time_until_tick > 0.0 {
            self.time_until_tick -= delta;
            if self.time_until_tick <= 0.0 {
                self.time_until_tick = rnd_value(MOON_MIN_TIME_BETWEEN_TICKS, MOON_MAX_TIME_BETWEEN_TICKS);
                return true;
            }
        }
        false
    }

    pub fn chance_to_shoot(&self) -> u32 {
        if self.anger.nearly_equal(0.0, 0.001) {
            0
        } else {
            (MOON_BASE_CHANCE_TO_SHOOT + 1.07f32.powf(0.9 * self.anger)) as u32
        }
    }

    pub fn should_shoot(&self) -> bool {
        let chance = self.chance_to_shoot();
        if chance == 0 {
            false
        } else {
            rnd_value(0, 99) < chance
        }
    }

    pub fn anger_color_offset(&self) -> u8 {
        let anger_level = self.chance_to_shoot();
        if anger_level == 0 {
            0
        } else if anger_level < 60 {
            3
        } else {
            6
        }
    }

}

pub struct MoonRock;

pub struct Pickupable {
    pub kind: PickupKind,
    pub pre_timer: f32,
}

pub struct Pickuper;

pub struct Particle;

pub struct IgnoresCollision;

pub struct IgnoresFriction;

pub struct Position(pub Vector2);

pub struct Velocity(pub Vector2);

#[derive(Debug, Copy, Clone, PartialEq)]
pub struct Force {
    pub force: Vector2,
    pub dissipation_factor: f32,
}

pub struct Forces {
    pub forces: Vec<Force>,
}

impl Forces {
    pub fn new() -> Self {
        Forces {
            forces: Vec::with_capacity(5),
        }
    }

    pub fn current_force(&self) -> Vector2 {
        let mut total_force = Vector2::ZERO;
        for force in self.forces.iter() {
            total_force += force.force;
        }
        total_force
    }

    pub fn add(&mut self, force: Vector2, dissipation_factor: f32) {
        self.forces.push(Force { force, dissipation_factor });
    }

    pub fn decay(&mut self) {
        for force in self.forces.iter_mut() {
            force.force *= force.dissipation_factor;
        }
        self.forces.retain(|f| !f.force.almost_zero(0.001));
    }
}

pub struct FacingAngle(pub f32);

pub struct Life(pub i32);

pub struct Shootable;

pub struct Projectile;

pub struct ShootAttack {
    pub time: f32,
}

impl ShootAttack {
    pub fn new() -> Self {
        ShootAttack {
            time: 0.0,
        }
    }
}

pub struct Channeler {
    pub kind: Option<ChannelKind>,
    pub time: f32,
    pub cycle: f32,
}

impl Channeler {
    pub fn new() -> Self {
        Channeler {
            time: 0.0,
            cycle: 0.0,
            kind: None,
        }
    }

    #[inline]
    pub fn is_channeling(&self) -> bool {
        self.time > 0.0 && self.kind.is_some()
    }

    #[inline]
    pub fn channel(&mut self, kind: ChannelKind) {
        self.time = CHANNEL_TIME;
        self.kind = Some(kind);
    }

    #[inline]
    pub fn stop(&mut self) {
        self.time = 0.0;
        self.cycle = 0.0;
        self.kind = None;
    }

    pub fn update(&mut self, delta: f32) -> bool {
        if self.time > 0.0 {
            self.cycle -= delta;
            if self.cycle <= 0.0 {
                self.cycle = CHANNEL_CYCLE_TIME;
            }

            self.time -= delta;
            if self.time <= 0.0 {
                self.cycle = 0.0;
                return true;
            }
        }
        false
    }
}

pub struct Channelable {
    pub time: f32,
    pub target: Vector2,
    pub allowed_kind: ChannelKind
}

impl Channelable {
    pub fn new(allowed_kind: ChannelKind) -> Self {
        Channelable {
            time: 0.0,
            target: Vector2::ZERO,
            allowed_kind,
        }
    }

    #[inline]
    pub fn is_channeled(&self) -> bool {
        self.time > 0.0
    }

    #[inline]
    pub fn is_channeled_by(&self, kind: ChannelKind) -> bool {
        self.is_channeled() && self.allowed_kind == kind
    }

    #[inline]
    pub fn stop(&mut self) {
        self.time = 0.0;
    }

    #[inline]
    pub fn refresh(&mut self, target: Vector2) {
        self.time = CHANNEL_CYCLE_TIME;
        self.target = target;
    }

    pub fn update(&mut self, delta: f32) -> bool {
        if self.time > 0.0 {
            self.time -= delta;
            if self.time <= 0.0 {
                return true;
            }
        }
        false
    }
}

pub struct FlungWhenShot;

pub struct Flung {
    pub max_distance: f32,
    pub distance: f32,
}

impl Flung {
    pub fn new(distance: f32) -> Self {
        Flung {
            max_distance: distance,
            distance: 0.0,
        }
    }

    #[inline]
    pub fn travelled(&self) -> f32 {
        self.distance / self.max_distance
    }

    pub fn update(&mut self, velocity: &Vector2, delta: f32) -> bool {
        let length = (*velocity * delta).length();
        self.distance += length;
        if self.distance >= self.max_distance {
            self.distance = self.max_distance;
            return true;
        }
        false
    }
}

pub struct Thrown;

pub struct Pushable;

pub struct Pusher {
    pub strength: f32,
    pub push_force_dissipation: f32,
}

impl Pusher {
    pub fn new() -> Self {
        Pusher {
            strength: DEFAULT_PUSH_STRENGTH,
            push_force_dissipation: DEFAULT_PUSH_DISSIPATION,
        }
    }
}

pub struct TargetPosition(pub Vector2);

pub struct Activity(pub EntityActivity);

pub struct AnimateByActivity(pub Rc<HashMap<EntityActivity, Rc<AnimationDef>>>);

pub struct RandomlyWalksAround {
    pub min_walk_time: f32,
    pub max_walk_time: f32,
    pub chance_to_move: u32,
    pub min_cooldown: f32,
    pub max_cooldown: f32,
    pub cooldown_timer: f32,
}

impl RandomlyWalksAround {
    pub fn new(min_walk_time: f32, max_walk_time: f32, chance_to_move: u32, min_cooldown: f32, max_cooldown: f32) -> Self {
        RandomlyWalksAround {
            min_walk_time,
            max_walk_time,
            chance_to_move,
            min_cooldown,
            max_cooldown,
            cooldown_timer: 0.0,
        }
    }

    pub fn should_start_walking(&self) -> bool {
        rnd_value(0, 100) < self.chance_to_move
    }
}

pub struct WalkConstantlyInDirection(pub f32);

pub struct WalkingTime(pub f32);

pub struct MovementSpeed(pub f32);

pub struct Bounds {
    pub width: u32,
    pub height: u32,
    pub radius: u32,
}

pub struct ParticleEmitter {
    pub kind: ParticleKind,
    pub time: f32,
    pub min_time: f32,
    pub max_time: f32,
    pub min_count: usize,
    pub max_count: usize,
}

impl ParticleEmitter {
    pub fn new(kind: ParticleKind, min_time: f32, max_time: f32, min_count: usize, max_count: usize) -> Self {
        let mut result = ParticleEmitter {
            kind,
            time: 0.0,
            min_time,
            max_time,
            min_count,
            max_count,
        };
        result.reset();
        result
    }

    pub fn update(&mut self, delta: f32) -> bool {
        if self.time > 0.0 {
            self.time -= delta;
            if self.time <= 0.0 {
                self.time = 0.0;
                return true;
            }
        }
        false
    }

    #[inline]
    pub fn reset(&mut self) {
        self.time = rnd_value(self.min_time, self.max_time);
    }

    #[inline]
    pub fn emit_count(&self) -> usize {
        rnd_value(self.min_count, self.max_count)
    }
}

pub struct LifeTime {
    pub time: f32,
    pub initial_time: f32
}

impl LifeTime {
    pub fn new(time: f32) -> Self {
        LifeTime {
            time,
            initial_time: time
        }
    }

    #[inline]
    pub fn life_percent(&self) -> f32 {
        self.time / self.initial_time
    }

    pub fn update(&mut self, delta: f32) -> bool {
        if self.time > 0.0 {
            self.time -= delta;
            if self.time <= 0.0 {
                self.time = 0.0;
                return true;
            }
        }
        false
    }
}

pub struct ColorByLifeTime(pub u8, pub u8, pub u8, pub u8, pub u8);

impl ColorByLifeTime {
    pub fn color(&self, life_time: f32) -> u8 {
        if life_time >= 0.8 {
            self.0
        } else if life_time >= 0.6 {
            self.1
        } else if life_time >= 0.4 {
            self.2
        } else if life_time >= 0.2 {
            self.3
        } else {
            self.4
        }
    }
}

pub struct Camera {
    pub x: i32,
    pub y: i32,
}

pub struct Transform {
    pub angle: f32,
    pub scale_x: f32,
    pub scale_y: f32,
}

pub struct ConstantRotation(pub f32);

pub enum FlickerMethod {
    Color(u8),
    OnOff,
}

pub struct TimedFlicker {
    pub method: FlickerMethod,
    pub flick: bool,
    pub pre_timer: Option<f32>,
    pub timer: f32,
}

impl TimedFlicker {
    pub fn new(timer: f32, method: FlickerMethod) -> Self {
        TimedFlicker {
            timer,
            method,
            pre_timer: None,
            flick: true
        }
    }

    pub fn new_with_pre_timer(timer: f32, pre_timer: f32, method: FlickerMethod) -> Self {
        TimedFlicker {
            timer,
            method,
            pre_timer: Some(pre_timer),
            flick: true,
        }
    }

    pub fn update(&mut self, delta: f32) {
        if let Some(mut pre_timer) = self.pre_timer {
            pre_timer -= delta;
            if pre_timer <= 0.0 {
                self.pre_timer = None;
            } else {
                self.pre_timer = Some(pre_timer)
            }
        } else {
            self.timer -= delta;
            self.flick = !self.flick;
        }
    }
}

pub struct RendersAboveTilemapOverlay;

pub struct FlipHorizontallyBasedOnDirection;

pub struct ColorOffset(pub u8);

pub struct Pixel(pub u8);

pub struct Sprite {
    pub atlas: Rc<BitmapAtlas>,
    pub index: usize,
    pub color: u8,
    pub flip_horizontally: bool,
    pub flip_vertically: bool,
}

#[derive(Debug, Copy, Clone, Eq, PartialEq, Hash)]
pub enum AnimationMethod {
    SpriteIndex,
    ColorOffset,
}

#[derive(Debug)]
pub struct AnimationDef {
    pub frames: &'static [usize],
    pub loops: bool,
    pub delay: f32,
    pub method: AnimationMethod,
}

impl AnimationDef {
    #[inline]
    pub fn new(frames: &'static [usize], loops: bool, delay: f32, method: AnimationMethod) -> Self {
        AnimationDef {
            frames,
            loops,
            delay,
            method,
        }
    }
}

#[derive(Debug)]
pub struct AnimationInstance {
    pub def: Rc<AnimationDef>,
    pub frame_index: usize,
    pub frame_timer: f32,
    pub complete: bool,
    pub delay_override: Option<f32>,
}

impl AnimationInstance {
    #[inline]
    pub fn from(def: Rc<AnimationDef>) -> Self {
        AnimationInstance {
            def,
            frame_index: 0,
            frame_timer: 0.0,
            complete: false,
            delay_override: None,
        }
    }

    #[inline]
    pub fn change_to(&mut self, def: Rc<AnimationDef>) {
        self.def = def;
        self.reset();
    }

    #[inline]
    pub fn reset(&mut self) {
        self.frame_index = 0;
        self.frame_timer = 0.0;
        self.complete = false;
    }
}

pub struct KillWhenAnimationFinishes;

#[derive(Debug, Copy, Clone, Eq, PartialEq, Hash)]
pub enum TallyEvent {
    PeopleSpawn,//
    PeopleFlung,//
    PeopleKill,//
    PeopleAbsorb,//
    AnimalSpawn,//
    AnimalKill,//
    AnimalThrow,//
    MoonRockSpawn,//
    MoonRockPlayerHit,//
    FourLeafCloverSpawn,//
    FourLeafCloverPickup,//
    HeartSpawn,//
    HeartPickup,//
}

pub struct Tally(pub HashMap<TallyEvent, usize>);

impl Tally {
    pub fn new() -> Self {
        Tally(HashMap::new())
    }
}