use libretrogd::entities::*;

use crate::components::*;
use crate::game::*;
use crate::logic::*;
use crate::spawning::spawn_animation_effect;

pub enum Event {
    Remove(EntityId),
    Killed(EntityId),
    SpawnRandomPerson,
    SpawnRandomAnimal,
    SpawnEffect {
        kind: EffectKind,
        x: i32,
        y: i32,
    },
    AnimationFinished(EntityId),
    TurnAndMove(EntityId, f32),
    MoveForward(EntityId),
    Shoot(EntityId),
    Channel {
        entity: EntityId,
        kind: ChannelKind
    },
    StopChanneling(EntityId),
    StopChanneled(EntityId),
    Hurt {
        victim: EntityId,
        source: Option<EntityId>,
        amount: u32,
    },
    Heal {
        victim: EntityId,
        amount: u32,
    },
    Shot {
        victim: EntityId,
        projectile: EntityId,
    },
    Channelled {
        channeler: EntityId,
        victim: EntityId,
    },
    ChannelledAtChanneler {
        channeled: EntityId,
        channeler: EntityId,
        kind: ChannelKind,
    },
    SetActivity {
        entity: EntityId,
        activity: EntityActivity
    },
    Pickup {
        picked_up_by: EntityId,
        picked_up: EntityId
    },
    Fling {
        victim: EntityId,
        is_from_moon_rock_impact: bool,
    },
    FinishedFlung(EntityId),
    TargetReached(EntityId),
    EmitParticle {
        source: EntityId,
        kind: ParticleKind,
        count: usize,
    },
    PlayerDied,
    Tally(TallyEvent),
}

pub fn event_listener(event: &Event, context: &mut Core) -> bool {
    match event {
        Event::Remove(entity) => {
            context.entities.remove_entity(*entity);
        },
        Event::Killed(entity) => {
            if context.entities.has_entity(*entity) {
                kill_entity(context, *entity);
            }
        },
        Event::SpawnRandomPerson => {
            try_spawn_new_person(context);
        },
        Event::SpawnRandomAnimal => {
            try_spawn_new_animal(context);
        },
        Event::SpawnEffect { kind, x, y } => {
            spawn_animation_effect(context, *kind, *x, *y);
        },
        Event::AnimationFinished(entity) => {
            if context.entities.has_entity(*entity) {
                // if the entity's 'attack' animation just finished, move them back to 'idle'
                let activities = context.entities.components::<Activity>();
                if let Some(activity) = activities.get(entity) {
                    if activity.0 == EntityActivity::Shooting {
                        drop(activities);
                        stop_attacking(context, *entity);
                    }
                }
            }
        },
        Event::TurnAndMove(entity, angle) => {
            if context.entities.has_entity(*entity) {
                turn_and_move_entity(context, *entity, *angle);
            }
        },
        Event::MoveForward(entity) => {
            if context.entities.has_entity(*entity) {
                move_entity_forward(context, *entity);
            }
        },
        Event::Shoot(entity) => {
            if context.entities.has_entity(*entity) {
                shoot(context, *entity);
            }
        },
        Event::Channel { entity, kind } => {
            if context.entities.has_entity(*entity) {
                channel(context, *entity, *kind);
            }
        }
        Event::StopChanneling(entity) => {
            if context.entities.has_entity(*entity) {
                stop_channeling(context, *entity);
            }
        }
        Event::StopChanneled(entity) => {
            if context.entities.has_entity(*entity) {
                stop_channeled(context, *entity);
            }
        }
        Event::Hurt { victim, source, amount } => {
            if context.entities.has_entity(*victim) {
                hurt_entity(context, *victim, *source, *amount);
            }
        },
        Event::Heal { victim, amount } => {
            if context.entities.has_entity(*victim) {
                heal_entity(context, *victim, *amount);
            }
        }
        Event::Shot { victim, projectile } => {
            if context.entities.has_entity(*victim) && context.entities.has_entity(*projectile) {
                handle_shot_entity(context, *victim, *projectile);
            }
        },
        Event::Channelled { channeler,  victim } => {
            if context.entities.has_entity(*victim) {
                channeled(context, *channeler, *victim);
            }
        },
        Event::ChannelledAtChanneler { channeled, channeler, kind } => {
            if context.entities.has_entity(*channeled) && context.entities.has_entity(*channeler) {
                match kind {
                    ChannelKind::Throw => throw_entity_at_moon(context, *channeled),
                    ChannelKind::Abosrb => hasten_absorbtion(context, *channeled),
                    _ => {},
                }
            }
        },
        Event::SetActivity { entity, activity } => {
            if context.entities.has_entity(*entity) {
                set_entity_activity(context, *entity, *activity);
            }
        },
        Event::Pickup { picked_up_by, picked_up } => {
            if context.entities.has_entity(*picked_up_by) && context.entities.has_entity(*picked_up) {
                pickup_entity(context, *picked_up_by, *picked_up);
            }
        },
        Event::Fling { victim, is_from_moon_rock_impact } => {
            if context.entities.has_entity(*victim) {
                fling_entity_randomly(context, *victim, *is_from_moon_rock_impact);
            }
        },
        Event::FinishedFlung(entity) => {
            if context.entities.has_entity(*entity) {
                entity_flung_finished(context, *entity);
            }
        }
        Event::TargetReached(entity) => {
            if context.entities.has_entity(*entity) {
                if context.entities.has_component::<MoonRock>(*entity) {
                    kill_entity(context, *entity);
                }
            }
        },
        Event::EmitParticle { source, kind, count } => {
            if context.entities.has_entity(*source) {
                emit_particle(context, *source, *kind, *count);
            }
        },
        Event::PlayerDied => {
            if let Some((_, world)) = context.entities.components_mut::<World>().single_mut() {
                world.time_to_game_over = Some(3.0);
            }
        }
        Event::Tally(event) => {
            tally_event(context, *event);
        },
        _ => return false
    }
    true
}
