use std::collections::HashMap;
use std::io::Cursor;
use std::path::Path;
use std::rc::Rc;

use anyhow::{Context, Result};
use libretrogd::audio::AudioBuffer;
use libretrogd::entities::*;
use libretrogd::events::*;
use libretrogd::graphics::*;
use libretrogd::system::*;

use crate::components::{AnimationDef, AnimationMethod};
use crate::events::*;
use crate::logic::*;
use crate::support::*;
use crate::tilemap::*;

static ASSET_FONT: &[u8] = include_bytes!("../assets/small.fnt");
static ASSET_TITLE_FONT: &[u8] = include_bytes!("../assets/dp.fnt");
static ASSET_PALETTE: &[u8] = include_bytes!("../assets/game.pal");
static ASSET_TILES: &[u8] = include_bytes!("../assets/tiles.pcx");
static ASSET_EMOJIS: &[u8] = include_bytes!("../assets/emojis.pcx");
static ASSET_TILEMAP_TITLE_SCREEN: &[u8] = include_bytes!("../assets/title_screen.map.json");
static ASSET_TILEMAP_GAME: &[u8] = include_bytes!("../assets/scene.map.json");
static ASSET_SOUND_ABSORB: &[u8] = include_bytes!("../assets/absorb.wav");
static ASSET_SOUND_ANIMAL_THROW_1: &[u8] = include_bytes!("../assets/animal_throw_1.wav");
static ASSET_SOUND_ANIMAL_THROW_2: &[u8] = include_bytes!("../assets/animal_throw_2.wav");
static ASSET_SOUND_ANIMAL_THROW_3: &[u8] = include_bytes!("../assets/animal_throw_3.wav");
static ASSET_SOUND_GROUND_IMPACT_1: &[u8] = include_bytes!("../assets/ground_impact_1.wav");
static ASSET_SOUND_GROUND_IMPACT_2: &[u8] = include_bytes!("../assets/ground_impact_2.wav");
static ASSET_SOUND_GROUND_IMPACT_3: &[u8] = include_bytes!("../assets/ground_impact_3.wav");
static ASSET_SOUND_EXPLOSION_1: &[u8] = include_bytes!("../assets/explosion_1.wav");
static ASSET_SOUND_EXPLOSION_2: &[u8] = include_bytes!("../assets/explosion_2.wav");
static ASSET_SOUND_EXPLOSION_3: &[u8] = include_bytes!("../assets/explosion_3.wav");
static ASSET_SOUND_HURT: &[u8] = include_bytes!("../assets/hurt.wav");
static ASSET_SOUND_MENU_MOVE: &[u8] = include_bytes!("../assets/menu_move.wav");
static ASSET_SOUND_MENU_SELECT: &[u8] = include_bytes!("../assets/menu_select.wav");
static ASSET_SOUND_MOON_ROCK_SHOOT: &[u8] = include_bytes!("../assets/moon_rock_shoot.wav");
static ASSET_SOUND_PICKUP_CLOVER: &[u8] = include_bytes!("../assets/pickup_clover.wav");
static ASSET_SOUND_PICKUP_HEART: &[u8] = include_bytes!("../assets/pickup_heart.wav");

pub struct Core {
    pub delta: f32,
    pub system: System,
    pub entities: Entities,
    pub event_publisher: EventPublisher<Event>,
    pub palette: Palette,
    pub fade_out_palette: Palette,
    pub font: BitmaskFont,
    pub title_font: BitmaskFont,
    pub tilemap: TileMap,
    pub game_tilemap: TileMap,
    pub title_screen_tilemap: TileMap,
    pub tiles: Rc<BitmapAtlas>,
    pub emojis: Rc<BitmapAtlas>,
    pub player_activity_states: Rc<HashMap<EntityActivity, Rc<AnimationDef>>>,
    pub moon_activity_states: Rc<HashMap<EntityActivity, Rc<AnimationDef>>>,
    pub particle_animations: Rc<HashMap<ParticleKind, Rc<AnimationDef>>>,
    pub effect_animations: Rc<HashMap<EffectKind, Rc<AnimationDef>>>,
    pub sound_absorb: Rc<AudioBuffer>,
    pub sound_animal_throw: Vec<Rc<AudioBuffer>>,
    pub sound_explosion: Vec<Rc<AudioBuffer>>,
    pub sound_ground_impact: Vec<Rc<AudioBuffer>>,
    pub sound_hurt: Rc<AudioBuffer>,
    pub sound_menu_move: Rc<AudioBuffer>,
    pub sound_menu_select: Rc<AudioBuffer>,
    pub sound_moon_rock_shoot: Rc<AudioBuffer>,
    pub sound_pickup_clover: Rc<AudioBuffer>,
    pub sound_pickup_heart: Rc<AudioBuffer>,
}

pub struct Game {
    pub core: Core,
    pub component_systems: ComponentSystems<Core, Core>,
    pub event_listeners: EventListeners<Event, Core>,
}

impl Game {
    pub fn new(mut system: System) -> Result<Self> {

        let font = load_font_bytes(&mut Cursor::new(ASSET_FONT), "small.fnt")?;
        let title_font = load_font_bytes(&mut Cursor::new(ASSET_TITLE_FONT), "dp.fnt")?;
        let palette = load_palette_bytes(&mut Cursor::new(ASSET_PALETTE), "game.pal")?;
        let tiles = load_bitmap_bytes_atlas_autogrid(&mut Cursor::new(ASSET_TILES), "tiles.pcx", TILE_WIDTH, TILE_HEIGHT)?;
        let emojis = load_bitmap_bytes_atlas_autogrid(&mut Cursor::new(ASSET_EMOJIS), "emojis.pcx", 16, 16)?;
        let game_tilemap = load_tilemap_bytes(&mut Cursor::new(ASSET_TILEMAP_GAME), "scene.map.json")?;
        let title_screen_tilemap = load_tilemap_bytes(&mut Cursor::new(ASSET_TILEMAP_TITLE_SCREEN), "title_screen.map.json")?;
        let sound_absorb = load_wav_bytes(&mut Cursor::new(ASSET_SOUND_ABSORB), "absorb.wav", system.audio.spec())?;
        let sound_animal_throw_1 = load_wav_bytes(&mut Cursor::new(ASSET_SOUND_ANIMAL_THROW_1), "animal_throw_1.wav", system.audio.spec())?;
        let sound_animal_throw_2 = load_wav_bytes(&mut Cursor::new(ASSET_SOUND_ANIMAL_THROW_2), "animal_throw_2.wav", system.audio.spec())?;
        let sound_animal_throw_3 = load_wav_bytes(&mut Cursor::new(ASSET_SOUND_ANIMAL_THROW_3), "animal_throw_3.wav", system.audio.spec())?;
        let sound_ground_impact_1 = load_wav_bytes(&mut Cursor::new(ASSET_SOUND_GROUND_IMPACT_1), "ground_impact_1.wav", system.audio.spec())?;
        let sound_ground_impact_2 = load_wav_bytes(&mut Cursor::new(ASSET_SOUND_GROUND_IMPACT_2), "ground_impact_2.wav", system.audio.spec())?;
        let sound_ground_impact_3 = load_wav_bytes(&mut Cursor::new(ASSET_SOUND_GROUND_IMPACT_3), "ground_impact_3.wav", system.audio.spec())?;
        let sound_explosion_1 = load_wav_bytes(&mut Cursor::new(ASSET_SOUND_EXPLOSION_1), "explosion_1.wav", system.audio.spec())?;
        let sound_explosion_2 = load_wav_bytes(&mut Cursor::new(ASSET_SOUND_EXPLOSION_2), "explosion_2.wav", system.audio.spec())?;
        let sound_explosion_3 = load_wav_bytes(&mut Cursor::new(ASSET_SOUND_EXPLOSION_3), "explosion_3.wav", system.audio.spec())?;
        let sound_hurt = load_wav_bytes(&mut Cursor::new(ASSET_SOUND_HURT), "hurt.wav", system.audio.spec())?;
        let sound_menu_move = load_wav_bytes(&mut Cursor::new(ASSET_SOUND_MENU_MOVE), "menu_move.wav", system.audio.spec())?;
        let sound_menu_select = load_wav_bytes(&mut Cursor::new(ASSET_SOUND_MENU_SELECT), "menu_select.wav", system.audio.spec())?;
        let sound_moon_rock_shoot = load_wav_bytes(&mut Cursor::new(ASSET_SOUND_MOON_ROCK_SHOOT), "moon_rock_shoot.wav", system.audio.spec())?;
        let sound_pickup_clover = load_wav_bytes(&mut Cursor::new(ASSET_SOUND_PICKUP_CLOVER), "pickup_clover.wav", system.audio.spec())?;
        let sound_pickup_heart = load_wav_bytes(&mut Cursor::new(ASSET_SOUND_PICKUP_HEART), "pickup_heart.wav", system.audio.spec())?;

        system.palette = palette.clone();

        let entities = Entities::new();
        let component_systems = ComponentSystems::new();
        let event_publisher = EventPublisher::new();
        let event_listeners = EventListeners::new();

        let player_activity_states = HashMap::from([
            (EntityActivity::Idle, Rc::new(AnimationDef::new(&[PLAYER_SPRITE_COLOR as usize], true, 1.0, AnimationMethod::ColorOffset))),
            (EntityActivity::Walking, Rc::new(AnimationDef::new(&[PLAYER_SPRITE_COLOR as usize], true, 1.0, AnimationMethod::ColorOffset))),
            (EntityActivity::Channeling(ChannelKind::Abosrb), Rc::new(AnimationDef::new(&[PLAYER_SPRITE_COLOR as usize, 100], true, 0.10, AnimationMethod::ColorOffset))),
            (EntityActivity::Channeling(ChannelKind::Throw), Rc::new(AnimationDef::new(&[PLAYER_SPRITE_COLOR as usize, 68], true, 0.10, AnimationMethod::ColorOffset))),
            (EntityActivity::Dead, Rc::new(AnimationDef::new(&[PLAYER_SPRITE_COLOR as usize], true, 0.0, AnimationMethod::ColorOffset))),
        ]);

        let moon_activity_states = HashMap::from([
            (EntityActivity::Idle, Rc::new(AnimationDef::new(&[240], true, 1.0, AnimationMethod::ColorOffset))),
            (EntityActivity::Walking, Rc::new(AnimationDef::new(&[240], true, 1.0, AnimationMethod::ColorOffset))),
            (EntityActivity::Shooting, Rc::new(AnimationDef::new(&[240, 241], false, 0.15, AnimationMethod::ColorOffset))),
            (EntityActivity::Dead, Rc::new(AnimationDef::new(&[240], true, 0.0, AnimationMethod::ColorOffset))),
        ]);

        let particle_animations = HashMap::from([
            (ParticleKind::Debris, Rc::new(AnimationDef::new(&[21, 20, 22, 23, 24], false, 0.15, AnimationMethod::ColorOffset))),
            (ParticleKind::FireyDebris, Rc::new(AnimationDef::new(&[21, 8, 5, 28, 27, 2], false, 0.15, AnimationMethod::ColorOffset))),
            (ParticleKind::RisingGlitter, Rc::new(AnimationDef::new(&[21, 20, 19, 18, 17, 16, 15], false, 0.15, AnimationMethod::ColorOffset))),
            (ParticleKind::RisingMagic, Rc::new(AnimationDef::new(&[9, 10, 11], false, 0.25, AnimationMethod::ColorOffset))),
        ]);

        let effect_animations = HashMap::from([
            (EffectKind::PowerUp, Rc::new(AnimationDef::new(&[172, 123, 482, 306], false, 0.07, AnimationMethod::SpriteIndex))),
            (EffectKind::Zap, Rc::new(AnimationDef::new(&[168, 167, 166], false, 0.15, AnimationMethod::SpriteIndex))),
        ]);

        let sound_animal_throw = vec![
            Rc::new(sound_animal_throw_1),
            Rc::new(sound_animal_throw_2),
            Rc::new(sound_animal_throw_3),
        ];
        let sound_explosion = vec![
            Rc::new(sound_explosion_1),
            Rc::new(sound_explosion_2),
            Rc::new(sound_explosion_3),
        ];
        let sound_ground_impact = vec![
            Rc::new(sound_ground_impact_1),
            Rc::new(sound_ground_impact_2),
            Rc::new(sound_ground_impact_3),
        ];

        Ok(Game {
            core: Core {
                delta: 0.0,
                system,
                entities,
                event_publisher,
                palette,
                fade_out_palette: Palette::new_with_default(0, 0, 0),
                font,
                title_font,
                tilemap: title_screen_tilemap.clone(),
                game_tilemap,
                title_screen_tilemap,
                tiles: Rc::new(tiles),
                emojis: Rc::new(emojis),
                player_activity_states: Rc::new(player_activity_states),
                moon_activity_states: Rc::new(moon_activity_states),
                particle_animations: Rc::new(particle_animations),
                effect_animations: Rc::new(effect_animations),
                sound_absorb: Rc::new(sound_absorb),
                sound_animal_throw,
                sound_explosion,
                sound_ground_impact,
                sound_hurt: Rc::new(sound_hurt),
                sound_menu_move: Rc::new(sound_menu_move),
                sound_menu_select: Rc::new(sound_menu_select),
                sound_moon_rock_shoot: Rc::new(sound_moon_rock_shoot),
                sound_pickup_clover: Rc::new(sound_pickup_clover),
                sound_pickup_heart: Rc::new(sound_pickup_heart),
            },
            component_systems,
            event_listeners,
        })
    }

    pub fn do_events(&mut self) {
        self.event_listeners.take_queue_from(&mut self.core.event_publisher);
        self.event_listeners.dispatch_queue(&mut self.core);
    }

    pub fn update_frame_delta(&mut self, last_ticks: u64) -> u64 {
        let ticks = self.core.system.ticks();
        let elapsed = ticks - last_ticks;
        self.core.delta = (elapsed as f64 / self.core.system.tick_frequency() as f64) as f32;
        ticks
    }
}
