use libretrogd::entities::*;
use libretrogd::math::*;
use libretrogd::SCREEN_WIDTH;
use libretrogd::utils::rnd_value;

use crate::components::*;
use crate::events::*;
use crate::game::*;
use crate::spawning::*;
use crate::support::*;
use crate::tilemap::*;

pub const GAME_TIME: f32 = 60.0 * 3.0;

pub const PLAYER_SPRITE: PeopleSprites = PeopleSprites::Shepard;
pub const PLAYER_SPRITE_COLOR: u8 = 156;

pub const MAX_LIFE: i32 = 6;

pub const FRICTION: f32 = 0.5;

pub const DEFAULT_PUSH_STRENGTH: f32 = 0.5;
pub const DEFAULT_PUSH_DISSIPATION: f32 = 0.1;

pub const HIT_KNOCKBACK_STRENGTH: f32 = 8.0;
pub const HIT_KNOCKBACK_DISSIPATION: f32 = 0.5;

pub const CHANNEL_DISTANCE: f32 = 40.0;
pub const CHANNEL_TIME: f32 = 0.25;
pub const CHANNEL_CYCLE_TIME: f32 = 0.25;
pub const CHANNELLED_FORCE_STRENGTH: f32 = 0.5;
pub const CHANNELLED_FORCE_DISSIPATION: f32 = 0.0;
pub const CHANNELLED_VELOCITY_FACTOR: f32 = 2.0;
pub const ABSORB_HASTEN_FACTOR: f32 = 1.1;

pub const PICKUP_PRE_TIMER: f32 = 0.5;
pub const PICKUP_ATTRACT_DISTANCE: f32 = 16.0;

pub const MAX_PERSONS: usize = 100;
pub const MAX_ANIMALS: usize = 3;

pub const PERSON_SPAWN_TIMER_MIN: f32 = 0.25;
pub const PERSON_SPAWN_TIMER_MAX: f32 = 1.0;

pub const ANIMAL_SPAWN_TIMER_MIN: f32 = 0.5;
pub const ANIMAL_SPAWN_TIMER_MAX: f32 = 20.0;
pub const ANIMAL_SPAWN_TIMER_PITY: f32 = 3.0;

pub const MOON_X: i32 = SCREEN_WIDTH as i32 / 2;
pub const MOON_Y: i32 = 20;

pub const MOON_ROCK_SPEED: f32 = 64.0;
pub const MOON_ROCK_IMPACT_RADIUS: u32 = 12;
pub const MOON_ROCK_DAMAGE: u32 = 1;

pub const FLUNG_SPEED: f32 = 12.0;
pub const FLUNG_MIN_DISTANCE: u32 = 32;
pub const FLUNG_MAX_DISTANCE: u32 = 64;

pub const THROW_SPEED: f32 = 80.0;

pub const ANIMAL_THROW_ANGER_AMOUNT: f32 = 20.0;
pub const MOON_MIN_TIME_BETWEEN_TICKS: f32 = 0.5;
pub const MOON_MAX_TIME_BETWEEN_TICKS: f32 = 1.0;
pub const MOON_BASE_CHANCE_TO_SHOOT: f32 = 20.0;
pub const MOON_MAX_ANGER: f32 = 100.0;

pub const DROP_ANYTHING_CHANCE: u32 = 50;

#[derive(Debug, Copy, Clone, Eq, PartialEq, Hash)]
pub enum MiscSprites {
    Moon = 1600,
    MoonRock = 332,
    FourLeafClover = 496,
    HeartFull = 165,
    HeartHalf = 164,
    HeartEmpty = 163,
    Clock = 147,
}

impl MiscSprites {
    pub const fn color(&self) -> u8 {
        use MiscSprites::*;
        match self {
            Moon => 48,
            MoonRock => 124,
            FourLeafClover => 63,
            HeartFull => 163,
            HeartHalf => 163,
            HeartEmpty => 163,
            Clock => 100,
        }
    }
}

#[derive(Debug, Copy, Clone, Eq, PartialEq, Hash)]
pub enum EffectKind {
    PowerUp,
    PickUp,
    Zap,
}

#[derive(Debug, Copy, Clone, Eq, PartialEq, Hash)]
pub enum ChannelKind {
    Abosrb,
    Throw,
}

impl ChannelKind {
    pub fn color(&self) -> u8 {
        use ChannelKind::*;
        match self {
            Abosrb => 8,
            Throw => 19,
        }
    }

    pub fn force_strength(&self) -> f32 {
        use ChannelKind::*;
        match self {
            Abosrb => 2.0,
            Throw => 0.5,
        }
    }

    pub fn distance(&self) -> f32 {
        use ChannelKind::*;
        match self {
            Abosrb => 48.0,
            Throw => 48.0,
        }
    }
}

#[derive(Debug, Copy, Clone, Eq, PartialEq, Hash)]
pub enum EntityActivity {
    Idle,
    Walking,
    Channeling(ChannelKind),
    Shooting,
    Dead,
}

#[derive(Debug, Copy, Clone, Eq, PartialEq, Hash)]
pub enum PickupKind {
    FourLeafClover,
    Heart,
}

impl PickupKind {
    pub fn new_random() -> Self {
        use PickupKind::*;
        match rnd_value(0, 1) {
            0 => FourLeafClover,
            _ => Heart,
        }
    }
}

#[derive(Debug, Copy, Clone, Eq, PartialEq, Hash)]
pub enum ParticleKind {
    Debris,
    FireyDebris,
    RisingGlitter,
    RisingMagic,
    Dust,
}

pub fn get_random_cardinal_direction_angle() -> f32 {
    match rnd_value(0, 3) {
        0 => UP,
        1 => DOWN,
        2 => LEFT,
        _ => RIGHT,
    }
}

pub fn circle_from_position_and_bounds(position: &Position, bounds: &Bounds) -> Circle {
    Circle::new(
        position.0.x as i32 + (bounds.width / 2) as i32,
        position.0.y as i32 + (bounds.height / 2) as i32,
        bounds.radius,
    )
}

pub fn rect_from_position_and_bounts(position: &Position, bounds: &Bounds) -> Rect {
    Rect::new(
        position.0.x as i32,
        position.0.y as i32,
        bounds.width,
        bounds.height
    )
}

pub fn get_entity_rect(entities: &Entities, entity: EntityId) -> Option<Rect> {
    let positions = entities.components::<Position>();
    let bounds = entities.components::<Bounds>();
    if let Some(position) = positions.get(&entity) {
        if let Some(bound) = bounds.get(&entity) {
            return Some(rect_from_position_and_bounts(position, bound))
        }
    }
    None
}

pub fn get_entity_circle(entities: &Entities, entity: EntityId) -> Option<Circle> {
    let positions = entities.components::<Position>();
    let bounds = entities.components::<Bounds>();
    if let Some(position) = positions.get(&entity) {
        if let Some(bound) = bounds.get(&entity) {
            return Some(circle_from_position_and_bounds(position, bound))
        }
    }
    None
}

pub fn get_entity_position(entities: &Entities, entity: EntityId) -> Option<Vector2> {
    let positions = entities.components::<Position>();
    if let Some(position) = positions.get(&entity) {
        Some(position.0)
    } else {
        None
    }
}

pub fn get_entity_xy(entities: &Entities, entity: EntityId) -> Option<(f32, f32)> {
    if let Some(position) = get_entity_position(entities, entity) {
        Some((position.x, position.y))
    } else {
        None
    }
}

pub fn get_entity_center_position(entities: &Entities, entity: EntityId) -> Option<Vector2> {
    let positions = entities.components::<Position>();
    let bounds = entities.components::<Bounds>();
    if let Some(position) = positions.get(&entity) {
        if let Some(bound) = bounds.get(&entity) {
            return Some(Vector2::new(
                position.0.x + (bound.width / 2) as f32,
                position.0.y + (bound.height / 2) as f32
            ));
        } else {
            return Some(position.0);
        }
    }
    None
}

pub fn get_entity_center_xy(entities: &Entities, entity: EntityId) -> Option<(f32, f32)> {
    if let Some(position) = get_entity_center_position(entities, entity) {
        Some((position.x, position.y))
    } else {
        None
    }
}

pub fn get_player_entity_id(entities: &Entities) -> Option<EntityId> {
    if let Some((entity_id, _)) = entities.components::<Player>().single() {
        Some(*entity_id)
    } else {
        None
    }
}

pub fn get_moon_entity_id(entities: &Entities) -> Option<EntityId> {
    if let Some((entity_id, _)) = entities.components::<Moon>().single() {
        Some(*entity_id)
    } else {
        None
    }
}

pub fn get_world_entity_id(entities: &Entities) -> Option<EntityId> {
    if let Some((entity_id, _)) = entities.components::<World>().single() {
        Some(*entity_id)
    } else {
        None
    }
}

pub fn get_channelable_kind(entities: &Entities, entity: EntityId) -> Option<ChannelKind> {
    let channelables = entities.components::<Channelable>();
    if let Some(channelable) = channelables.get(&entity) {
        Some(channelable.allowed_kind)
    } else {
        None
    }
}

pub fn is_channeled(entities: &Entities, entity: EntityId, kind: ChannelKind) -> bool {
    let channelables = entities.components::<Channelable>();
    if let Some(channelable) = channelables.get(&entity) {
        channelable.is_channeled_by(kind)
    } else {
        false
    }
}

pub fn move_entity_with_collision(
    position: &mut Position,
    bounds: &Bounds,
    velocity: Option<&Velocity>,
    forces: Option<&Forces>,
    map: &TileMap,
    delta: f32
) -> bool {
    const NUM_STEPS: usize = 2;

    let mut collided = false;

    // apply entity velocity + force (if any/either) and exit early with no collision if this entity
    // has no movement ... no need to check for collisions in such a case
    let mut step_velocity = Vector2::ZERO;
    if let Some(velocity) = velocity {
        step_velocity += velocity.0 * delta;
    }
    if let Some(forces) = forces {
        step_velocity += forces.current_force();
    }
    if step_velocity.nearly_equal(Vector2::ZERO, 0.00001) {
        return collided;
    }

    // entity is actually moving, so check for collisions and move accordingly
    step_velocity *= 1.0 / NUM_STEPS as f32;
    for _ in 0..NUM_STEPS {
        let old_position = position.0;

        position.0.x += step_velocity.x;
        if map.is_colliding(&Rect::new(position.0.x as i32, position.0.y as i32, bounds.width, bounds.height)) {
            collided = true;
            position.0.x = old_position.x;
        }

        position.0.y += step_velocity.y;
        if map.is_colliding(&Rect::new(position.0.x as i32, position.0.y as i32, bounds.width, bounds.height)) {
            collided = true;
            position.0.y = old_position.y;
        }
    }

    collided
}

pub fn move_entity_forward(context: &mut Core, entity: EntityId) {
    let mut velocities = context.entities.components_mut::<Velocity>();
    let facing_angles = context.entities.components::<FacingAngle>();
    let movement_speeds = context.entities.components::<MovementSpeed>();

    let velocity = velocities.get_mut(&entity).unwrap();
    let facing_angle = facing_angles.get(&entity).unwrap();
    let movement_speed = movement_speeds.get(&entity).unwrap();

    let movement = Vector2::from_angle(facing_angle.0) * movement_speed.0;

    velocity.0 += movement;
}

pub fn turn_and_move_entity(context: &mut Core, entity: EntityId, angle: f32) {
    // can this entity currently move at all?
    let activities = context.entities.components::<Activity>();
    if let Some(activity) = activities.get(&entity) {
        if activity.0 != EntityActivity::Idle && activity.0 != EntityActivity::Walking {
            return;
        }
    }
    drop(activities);

    // make the entity face in the direction specified
    let mut facing_angles = context.entities.components_mut::<FacingAngle>();
    if let Some(facing_angle) = facing_angles.get_mut(&entity) {
        facing_angle.0 = angle;
    }
    drop(facing_angles);

    move_entity_forward(context, entity);
}

pub fn set_entity_activity(context: &mut Core, entity: EntityId, new_activity: EntityActivity) {
    let mut activities = context.entities.components_mut::<Activity>();
    if let Some(activity) = activities.get_mut(&entity) {
        // only change the activity and the entities animation, if appropriate, if we are actually
        // applying an activity change from what it was before. for animations, this is important
        // so that we're not (maybe) constantly resetting the animation
        if activity.0 != new_activity {
            activity.0 = new_activity;

            let animate_by_activitys = context.entities.components::<AnimateByActivity>();
            if let Some(animate_by_activity) = animate_by_activitys.get(&entity) {
                if let Some(new_animation_def) = animate_by_activity.0.get(&new_activity) {
                    let mut animations = context.entities.components_mut::<AnimationInstance>();
                    if let Some(animation) = animations.get_mut(&entity) {
                        animation.change_to(new_animation_def.clone());
                    }
                }
            }
        }
    }
}

pub fn channel(context: &mut Core, entity: EntityId, channel_kind: ChannelKind) {
    let activities = context.entities.components::<Activity>();
    let mut channeler = context.entities.components_mut::<Channeler>();
    if let Some(activity) = activities.get(&entity) {
        match activity.0 {
            EntityActivity::Idle | EntityActivity::Walking => {
                // not already channeling, so we need to start it and change the entity activity
                if let Some(channeler) = channeler.get_mut(&entity) {
                    channeler.channel(channel_kind);
                }
                drop(activities);
                drop(channeler);
                set_entity_activity(context, entity, EntityActivity::Channeling(channel_kind));
            },
            EntityActivity::Channeling(kind) if kind == channel_kind => {
                // just refresh the existing channel effect
                if let Some(channeler) = channeler.get_mut(&entity) {
                    channeler.channel(channel_kind);
                }
            }
            _ => {},
        }
    }
}

pub fn stop_channeling(context: &mut Core, entity: EntityId) {
    let mut channelers = context.entities.components_mut::<Channeler>();
    if let Some(channeler) = channelers.get_mut(&entity) {
        channeler.stop();
        drop(channelers);
        set_entity_activity(context, entity, EntityActivity::Idle);
    }
}

pub fn channeled(context: &mut Core, channeler: EntityId, victim: EntityId) {
    if get_channelable_kind(&context.entities, victim) == Some(ChannelKind::Abosrb) {
        let mut velocities = context.entities.components_mut::<Velocity>();
        let mut velocity = velocities.get_mut(&victim).unwrap();
        velocity.0 *= CHANNELLED_VELOCITY_FACTOR;
    }
}

pub fn hasten_absorbtion(context: &mut Core, victim: EntityId) {
    let mut flungs = context.entities.components_mut::<Flung>();
    if let Some(mut flung) = flungs.get_mut(&victim) {
        flung.distance *= ABSORB_HASTEN_FACTOR;
    }
}

pub fn stop_channeled(context: &mut Core, entity: EntityId) {
    let mut channelables = context.entities.components_mut::<Channelable>();
    if let Some(channelable) = channelables.get_mut(&entity) {
        channelable.stop();
        drop(channelables);
        // TODO more things?
    }
}

fn shoot_moon_rock_at_player(context: &mut Core, moon_entity: EntityId) {
    let source_position;
    let target_position;

    if let Some(player) = get_player_entity_id(&context.entities) {
        let positions = context.entities.components::<Position>();
        let bounds = context.entities.components::<Bounds>();

        source_position = positions.get(&moon_entity).unwrap().0;
        let player_position = positions.get(&player).unwrap();
        let player_bounds = bounds.get(&player).unwrap();

        target_position = Vector2::new(
            player_position.0.x + rnd_value(-(player_bounds.width as i32), (player_bounds.width * 2) as i32) as f32,
            player_position.0.y + rnd_value(-(player_bounds.height as i32), (player_bounds.height * 2) as i32) as f32,
        );

    } else {
        return;
    }

    context.event_publisher.queue(Event::Tally(TallyEvent::MoonRockSpawn));
    play_sound_moon_rock_shoot(context);
    new_moon_rock_entity(context, source_position.x as i32, source_position.y as i32, target_position, MOON_ROCK_SPEED);
}

pub fn shoot(context: &mut Core, entity: EntityId) {
    let activities = context.entities.components::<Activity>();
    if let Some(activity) = activities.get(&entity) {
        match activity.0 {
            EntityActivity::Idle | EntityActivity::Walking => {
                drop(activities);
                set_entity_activity(context, entity, EntityActivity::Shooting);
                if context.entities.has_component::<Moon>(entity) {
                    shoot_moon_rock_at_player(context, entity);
                }
            },
            _ => {},
        }
    }
}

pub fn stop_attacking(context: &mut Core, entity: EntityId) {
    set_entity_activity(context, entity, EntityActivity::Idle);
}

fn convert_to_flung_entity(context: &mut Core, victim: EntityId, velocity: Vector2, distance: f32) -> EntityId {
    let sprite_index;
    let color;
    {
        let sprites = context.entities.components::<Sprite>();
        let sprite = sprites.get(&victim).unwrap();
        sprite_index = sprite.index;
        color = sprite.color;
    }
    let (x, y) = get_entity_xy(&context.entities, victim).unwrap();

    let id = new_flung_entity(context, x as i32, y as i32, sprite_index, color, velocity, distance);
    context.entities.remove_entity(victim);
    id
}

pub fn fling_entity_randomly(context: &mut Core, victim: EntityId, is_from_moon_rock_hit: bool) -> EntityId {
    context.event_publisher.queue(Event::Tally(TallyEvent::PeopleFlung));

    let angle = rnd_value(RADIANS_0, RADIANS_360);
    let distance = rnd_value(FLUNG_MIN_DISTANCE, FLUNG_MAX_DISTANCE) as f32;
    let speed = FLUNG_SPEED;
    let velocity = Vector2::from_angle(angle) * speed;

    let id = convert_to_flung_entity(context, victim, velocity, distance);

    if is_from_moon_rock_hit {
        context.entities.add_component(id, Channelable::new(ChannelKind::Abosrb));
    }

    id
}

pub fn fling_entity(context: &mut Core, victim: EntityId, target: Vector2, velocity: Vector2, is_from_moon_rock_hit: bool) -> EntityId {
    let victim_position = get_entity_position(&context.entities, victim).unwrap();
    let distance = victim_position.distance(&target);

    let id = convert_to_flung_entity(context, victim, velocity, distance);

    if is_from_moon_rock_hit {
        context.entities.add_component(id, Channelable::new(ChannelKind::Abosrb));
    }

    id
}

pub fn absorb_entity(context: &mut Core, victim: EntityId) {
    let (x, y) = get_entity_xy(&context.entities, victim).unwrap();

    context.event_publisher.queue(Event::SpawnEffect {
        kind: EffectKind::Zap,
        x: x as i32,
        y: y as i32,
    });
    new_particle_emitter(
        context,
        x as i32,
        y as i32,
        (ChannelKind::Abosrb.distance() as u32) / 2,
        1.0,
        ParticleKind::RisingGlitter,
        0.1,
        0.5,
        4,
        8
    );
    context.event_publisher.queue(Event::Tally(TallyEvent::PeopleAbsorb));
    play_sound_absorb(context);

    maybe_spawn_pickups_from_entity(context, victim);

    context.entities.remove_entity(victim);
}

pub fn animal_hit_moon(context: &mut Core, victim: EntityId) {
    if let Some(region) = get_entity_rect(&context.entities, victim) {
        let (x, y) = get_entity_center_xy(&context.entities, victim).unwrap();
        spawn_debris_circle(
            context,
            x as i32, y as i32,
            16,
            20.0,
        );
        spawn_debris_cloud(context, region);
        play_sound_ground_impact(context);
    }
    context.entities.remove_entity(victim);

    if let Some((_, mut moon)) = context.entities.components_mut::<Moon>().single_mut() {
        moon.anger += ANIMAL_THROW_ANGER_AMOUNT;
    }
}

pub fn entity_flung_finished(context: &mut Core, victim: EntityId) {
    if is_channeled(&context.entities, victim, ChannelKind::Abosrb) {
        absorb_entity(context, victim);
        return;
    }

    if let Some(moon_entity_id) = get_moon_entity_id(&context.entities) {
        let moon_circle = get_entity_circle(&context.entities, moon_entity_id).unwrap();
        let victim_circle = get_entity_circle(&context.entities, victim).unwrap();
        if moon_circle.overlaps(&victim_circle) {
            animal_hit_moon(context, victim);
            return;
        }
    }

    kill_entity(context, victim);
}

pub fn handle_shot_entity(context: &mut Core, victim: EntityId, projectile: EntityId) {
    if context.entities.has_component::<FlungWhenShot>(victim) {
        let id = fling_entity_randomly(context, victim, context.entities.has_component::<MoonRock>(projectile));
    } else {
        if context.entities.has_component::<Player>(victim) {
            context.event_publisher.queue(Event::Tally(TallyEvent::MoonRockPlayerHit));
            play_sound_hurt(context);
        }
        hurt_entity(context, victim, Some(projectile), MOON_ROCK_DAMAGE);
    }
}

pub fn moon_rock_impact_region_at(context: &mut Core, region: Circle) {
    if let Some(flingables) = context.entities.components::<FlungWhenShot>() {
        for (entity, _) in flingables.iter() {
            if let Some(potential_victim_region) = get_entity_circle(&context.entities, *entity) {
                if potential_victim_region.overlaps(&region) {
                    context.event_publisher.queue(Event::Fling {
                        victim: *entity,
                        is_from_moon_rock_impact: true
                    });
                }
            }
        }
    }
}

pub fn throw_entity(context: &mut Core, victim: EntityId, target: EntityId) {
    let target_position = get_entity_center_position(&context.entities, target).unwrap();
    let victim_position = get_entity_position(&context.entities, victim).unwrap();

    let angle = victim_position.angle_between(&target_position);
    let speed = THROW_SPEED;
    let velocity = Vector2::from_angle(angle) * speed;

    if context.entities.has_component::<Animal>(victim) {
        play_sound_animal_throw(context);
        context.event_publisher.queue(Event::Tally(TallyEvent::AnimalThrow));
    }
    let flung_entity = fling_entity(context, victim, target_position, velocity, false);
    context.entities.add_component(flung_entity, ParticleEmitter::new(ParticleKind::Debris, 0.1, 0.15, 2, 5));
    context.entities.add_component(flung_entity, Thrown);
}

pub fn throw_entity_at_moon(context: &mut Core, victim: EntityId) {
    if let Some(moon_entity) = get_moon_entity_id(&context.entities) {
        throw_entity(context, victim, moon_entity);
    }
}

pub fn try_spawn_new_person(context: &mut Core) -> Option<EntityId> {
    if context.entities.components::<Person>().len() >= MAX_PERSONS {
        None
    } else {
        context.event_publisher.queue(Event::Tally(TallyEvent::PeopleSpawn));
        spawn_random_person(context)
    }
}

pub fn try_spawn_new_animal(context: &mut Core) -> Option<EntityId> {
    if context.entities.components::<Animal>().len() >= MAX_ANIMALS {
        None
    } else {
        context.event_publisher.queue(Event::Tally(TallyEvent::AnimalSpawn));
        spawn_random_animal(context)
    }
}

pub fn emit_particle(context: &mut Core, entity: EntityId, kind: ParticleKind, count: usize) {
    if let Some(rect) = get_entity_rect(&context.entities, entity) {
        use ParticleKind::*;
        match kind {
            Debris | FireyDebris => {
                let angle;
                let velocities = context.entities.components::<Velocity>().unwrap();
                if let Some(velocity) = velocities.get(&entity) {
                    angle = velocity.0.angle();
                } else {
                    panic!("TODO");
                }
                drop(velocities);

                for _ in 0..count {
                    let x = rnd_value(rect.x, rect.right());
                    let y = rnd_value(rect.y, rect.bottom());
                    if kind == Debris {
                        new_debris_particle(context, x, y, Vector2::from_angle(angle) * 8.0, None);
                    } else if kind == FireyDebris {
                        new_firey_debris_particle(context, x, y, Vector2::from_angle(angle) * 8.0, None);
                    }
                }
            },
            RisingGlitter => {
                for _ in 0..count {
                    let x = rnd_value(rect.x, rect.right());
                    let y = rnd_value(rect.y, rect.bottom());
                    new_rising_glitter_particle(context, x, y, Vector2::from_angle(UP) * 16.0, None);
                }
            },
            RisingMagic => {
                for _ in 0..count {
                    let x = rnd_value(rect.x, rect.right());
                    let y = rnd_value(rect.y, rect.bottom());
                    new_rising_magic_particle(context, x, y, Vector2::from_angle(UP) * 16.0, None);
                }
            }
            Dust => {},
        }
    }
}

pub fn add_knockback_force_to(context: &mut Core, victim: EntityId, source: EntityId) {
    let mut forces = context.entities.components_mut::<Forces>();
    if let Some(mut force) = forces.get_mut(&victim) {
        let source_position = get_entity_position(&context.entities, source).unwrap();
        let victim_position = get_entity_position(&context.entities, victim).unwrap();
        let knockback_direction = (victim_position - source_position).normalize();
        force.add(knockback_direction * HIT_KNOCKBACK_STRENGTH, HIT_KNOCKBACK_DISSIPATION);
    }
}

pub fn hurt_entity(context: &mut Core, victim: EntityId, source: Option<EntityId>, amount: u32) {
    let mut lifes = context.entities.components_mut::<Life>();
    let is_dead;
    if let Some(life) = lifes.get_mut(&victim) {
        life.0 -= amount as i32;
        is_dead = life.0 <= 0;
    } else {
        is_dead = false;
    }
    drop(lifes);

    if context.entities.has_component::<Player>(victim) ||
        context.entities.has_component::<Animal>(victim) ||
        context.entities.has_component::<Person>(victim) {
        spawn_hurt_blood_splatter_effect_at(context, victim);
    }

    if is_dead {
        kill_entity(context, victim);
    } else {
        context.entities.add_component(victim, TimedFlicker::new(1.0, FlickerMethod::OnOff));
        if let Some(source) = source {
            add_knockback_force_to(context, victim, source);
        }
    }
}

pub fn heal_entity(context: &mut Core, victim: EntityId, amount: u32) {
    let mut lifes = context.entities.components_mut::<Life>();
    if let Some(life) = lifes.get_mut(&victim) {
        life.0 += amount as i32;
        if life.0 > MAX_LIFE {
            life.0 = MAX_LIFE;
        }
    }
    drop(lifes);
    spawn_healed_effect_at(context, victim);
}

pub fn kill_entity(context: &mut Core, victim: EntityId) {
    if context.entities.has_component::<Flung>(victim) {
        context.event_publisher.queue(Event::Tally(TallyEvent::PeopleKill));
        play_sound_ground_impact(context);
        spawn_debris_cloud_at(context, victim);
    } else if context.entities.has_component::<MoonRock>(victim) {
        if let Some(region) = get_entity_rect(&context.entities, victim) {
            let (x, y) = get_entity_center_xy(&context.entities, victim).unwrap();
            spawn_firey_sparks_circle(
                context,
                x as i32, y as i32,
                16,
                20.0,
            );
            spawn_firey_debris_cloud(context, region);
            moon_rock_impact_region_at(context, Circle::new(x as i32, y as i32, MOON_ROCK_IMPACT_RADIUS));
            play_sound_explosion(context);
        }
    } else if context.entities.has_component::<Animal>(victim) {
        context.event_publisher.queue(Event::Tally(TallyEvent::AnimalKill));
        play_sound_ground_impact(context);
        spawn_death_blood_splatter_effect_at(context, victim);
    } else if context.entities.has_component::<Player>(victim) {
        spawn_death_blood_splatter_effect_at(context, victim);
        context.event_publisher.queue(Event::PlayerDied);

    }
    context.entities.remove_entity(victim);
}

pub fn pickup_entity(context: &mut Core, picked_up_by: EntityId, picked_up: EntityId) {
    let (x, y) = get_entity_xy(&context.entities, picked_up).unwrap();
    let kind;
    {
        let pickupables = context.entities.components::<Pickupable>();
        kind = pickupables.get(&picked_up).unwrap().kind;
    }

    spawn_animation_effect(context, EffectKind::PowerUp, x as i32, y as i32);
    context.entities.remove_entity(picked_up);

    match kind {
        PickupKind::FourLeafClover => {
            context.event_publisher.queue(Event::Tally(TallyEvent::FourLeafCloverPickup));
            play_sound_pickup_clover(context);
        },
        PickupKind::Heart => {
            context.event_publisher.queue(Event::Tally(TallyEvent::HeartPickup));
            play_sound_pickup_heart(context);
            heal_entity(context, picked_up_by, 2);
        }
    }
}

pub fn tally_event(context: &mut Core, event: TallyEvent) {
    let worlds = context.entities.components::<World>();
    let mut tallys = context.entities.components_mut::<Tally>();

    if let Some((world_entity_id, _)) = worlds.single() {
        if let Some(mut tally) = tallys.get_mut(world_entity_id) {
            if let Some(value) = tally.0.get_mut(&event) {
                *value += 1;
            } else {
                tally.0.insert(event, 1);
            }
        }
    }
}