use std::rc::Rc;

use libretrogd::entities::*;
use libretrogd::math::*;
use libretrogd::SCREEN_WIDTH;
use libretrogd::utils::rnd_value;

use crate::components::*;
use crate::events::Event;
use crate::game::*;
use crate::logic::*;
use crate::tilemap::*;

#[derive(Debug, Copy, Clone, Eq, PartialEq, Hash)]
pub enum AnimalSprites {
    Goose = 178,
    Chicken = 179,
    Cat = 180,
    Dog = 181,
    Cow = 182,
    Horse = 183,
    Sheep = 184,
    Pig = 185,
    //Snake = 186,
    //Deer = 187,
    //Stag = 188,
    //Boar = 189,
   // Llama = 190,
    //Alpaca = 191,
    //Gorilla = 192,
    //Monkey = 193,
    //Turkey = 194,
    Rabbit = 195,
    //Fox = 196,
    //Turtle = 197,
}

impl AnimalSprites {
    pub fn rnd() -> Self {
        use AnimalSprites::*;
        match rnd_value(0, 8) {
            0 => Goose,
            1 => Chicken,
            2 => Cat,
            3 => Dog,
            4 => Cow,
            5 => Horse,
            6 => Sheep,
            7 => Pig,
            _ => Rabbit,
        }
    }

    pub const fn color(&self) -> u8 {
        use AnimalSprites::*;
        match self {
            Goose => 48,
            Chicken => 100,
            Cat => 0,
            Dog => 74,
            Cow => 48,
            Horse => 54,
            Sheep => 150,
            Pig => 153,
            Rabbit => 134,
        }
    }
}

#[derive(Debug, Copy, Clone, Eq, PartialEq, Hash)]
pub enum PeopleSprites {
    //Warrior = 198,
    //Dwarf = 199,
    Elf = 200,
    Halfling = 201,
    //MagicUser = 202,
    Bard = 203,
    VillagerA = 204,
    VillagerB = 205,
    VillagerC = 206,
    Farmer = 207,
    Worker = 208,
    Karateka = 209,
    Ninja = 210,
    Priest = 211,
    Pirate = 212,
    Shepard = 213,
    Guard = 214,
    HighPriest = 215,
    StrongVillager = 216,
}

impl PeopleSprites {
    pub fn rnd() -> Self {
        use PeopleSprites::*;
        match rnd_value(0, 15) {
            0 => Elf,
            1 => Halfling,
            2 => Bard,
            3 => VillagerA,
            4 => VillagerB,
            5 => VillagerC,
            6 => Farmer,
            7 => Worker,
            8 => Karateka,
            9 => Ninja,
            10 => Priest,
            11 => Pirate,
            12 => Shepard,
            13 => Guard,
            14 => HighPriest,
            _ => StrongVillager,
        }
    }

    pub const fn color(&self) -> u8 {
        use PeopleSprites::*;
        match self {
            Elf => 122,
            Halfling => 116,
            Bard => 128,
            VillagerA => 131,
            VillagerB => 106,
            VillagerC => 71,
            Farmer => 57,
            Worker => 109,
            Karateka => 134,
            Ninja => 137,
            Priest => 97,
            Pirate => 140,
            Shepard => 160,
            Guard => 60,
            HighPriest => 144,
            StrongVillager => 147,
        }
    }
}

#[derive(Debug, Copy, Clone, Eq, PartialEq, Hash)]
pub enum WalkAroundType {
    Lazy,
    Normal,
    Hyperactive,
}

impl WalkAroundType {
    pub fn rnd() -> Self {
        use WalkAroundType::*;
        match rnd_value(0, 2) {
            0 => Lazy,
            1 => Normal,
            _ => Hyperactive,
        }
    }
}

pub fn new_world_entity(context: &mut Core, camera_x: i32, camera_y: i32, time: Option<f32>) -> EntityId {
    let id = context.entities.new_entity();
    context.entities.add_component(id, World::new(time));
    context.entities.add_component(id, Camera { x: camera_x, y: camera_y });
    context.entities.add_component(id, Tally::new());
    id
}

pub fn new_player_entity(context: &mut Core, x: i32, y: i32) -> EntityId {
    let id = context.entities.new_entity();

    let activity = EntityActivity::Idle;
    let animate_by_activity = AnimateByActivity(context.player_activity_states.clone());
    let animation = AnimationInstance::from(animate_by_activity.0.get(&activity).unwrap().clone());

    let people_sprite = PLAYER_SPRITE;

    let sprite = Sprite {
        atlas: context.tiles.clone(),
        index: people_sprite as usize,
        color: people_sprite.color(),
        flip_horizontally: false,
        flip_vertically: false,
    };

    context.entities.add_component(id, Player);
    context.entities.add_component(id, Position(Vector2::new(x as f32, y as f32)));
    context.entities.add_component(id, Velocity(Vector2::ZERO));
    context.entities.add_component(id, Forces::new());
    context.entities.add_component(id, Bounds { width: TILE_WIDTH, height: TILE_HEIGHT, radius: TILE_WIDTH / 2});
    context.entities.add_component(id, FacingAngle(0.0));
    context.entities.add_component(id, FlipHorizontallyBasedOnDirection);
    context.entities.add_component(id, sprite);
    context.entities.add_component(id, Activity(activity));
    context.entities.add_component(id, animate_by_activity);
    context.entities.add_component(id, animation);
    context.entities.add_component(id, MovementSpeed(16.0));
    context.entities.add_component(id, Pusher::new());
    context.entities.add_component(id, Pushable);
    context.entities.add_component(id, Channeler::new());
    context.entities.add_component(id, Life(6));
    context.entities.add_component(id, Shootable);
    context.entities.add_component(id, Pickuper);

    id
}

pub fn new_moon_entity(context: &mut Core, x: i32, y: i32) -> EntityId {
    let id = context.entities.new_entity();

    let activity = EntityActivity::Idle;
    let animate_by_activity = AnimateByActivity(context.moon_activity_states.clone());
    let animation = AnimationInstance::from(animate_by_activity.0.get(&activity).unwrap().clone());

    let sprite = Sprite {
        atlas: context.tiles.clone(),
        index: MiscSprites::Moon as usize,
        color: animation.def.frames[0] as u8,
        flip_horizontally: false,
        flip_vertically: false,
    };

    context.entities.add_component(id, Moon::new());
    context.entities.add_component(id, Position(Vector2::new(x as f32, y as f32)));
    context.entities.add_component(id, Velocity(Vector2::ZERO));
    context.entities.add_component(id, Bounds { width: TILE_WIDTH, height: TILE_HEIGHT, radius: TILE_WIDTH / 2});
    context.entities.add_component(id, sprite);
    context.entities.add_component(id, Activity(activity));
    context.entities.add_component(id, animate_by_activity);
    context.entities.add_component(id, animation);
    context.entities.add_component(id, ShootAttack::new());
    context.entities.add_component(id, IgnoresCollision);
    context.entities.add_component(id, IgnoresFriction);
    context.entities.add_component(id, ColorOffset(0));

    id
}

pub fn new_moon_rock_entity(context: &mut Core, x: i32, y: i32, target: Vector2, speed: f32) -> EntityId {
    let id = context.entities.new_entity();

    let sprite = Sprite {
        atlas: context.tiles.clone(),
        index: MiscSprites::MoonRock as usize,
        color: MiscSprites::MoonRock.color(),
        flip_horizontally: false,
        flip_vertically: false,
    };

    let position = Vector2::new(x as f32, y as f32);
    let velocity = (target - position).normalize() * speed;

    const SIZE: u32 = TILE_WIDTH;

    context.entities.add_component(id, MoonRock);
    context.entities.add_component(id, Projectile);
    context.entities.add_component(id, Position(position));
    context.entities.add_component(id, Velocity(velocity));
    context.entities.add_component(id, Bounds { width: SIZE, height: SIZE, radius: SIZE / 2});
    context.entities.add_component(id, sprite);
    context.entities.add_component(id, IgnoresCollision);
    context.entities.add_component(id, IgnoresFriction);
    context.entities.add_component(id, TargetPosition(target));
    context.entities.add_component(id, Life(1));
    context.entities.add_component(id, RendersAboveTilemapOverlay);
    context.entities.add_component(id, Transform { angle: 0.0, scale_x: 0.9, scale_y: 0.9 });
    context.entities.add_component(id, ConstantRotation(10.0f32.to_radians()));
    context.entities.add_component(id, ParticleEmitter::new(ParticleKind::Debris, 0.1, 0.25, 1, 3));

    id
}

pub fn new_person_entity(
    context: &mut Core,
    x: i32,
    y: i32,
    person_sprite: PeopleSprites,
    walk_around_type: WalkAroundType,
) -> EntityId {
    use WalkAroundType::*;

    let id = context.entities.new_entity();

    let activity = EntityActivity::Idle;

    let sprite = Sprite {
        atlas: context.tiles.clone(),
        index: person_sprite as usize,
        color: person_sprite.color(),
        flip_horizontally: false,
        flip_vertically: false,
    };

    let walk_direction = if x < (SCREEN_WIDTH / 2) as i32 {
        RADIANS_0
    } else {
        RADIANS_180
    };

    context.entities.add_component(id, Person);
    context.entities.add_component(id, Position(Vector2::new(x as f32, y as f32)));
    context.entities.add_component(id, Velocity(Vector2::ZERO));
    context.entities.add_component(id, Forces::new());
    context.entities.add_component(id, Bounds { width: TILE_WIDTH, height: TILE_HEIGHT, radius: TILE_WIDTH / 2});
    context.entities.add_component(id, FacingAngle(0.0));
    context.entities.add_component(id, FlipHorizontallyBasedOnDirection);
    context.entities.add_component(id, sprite);
    context.entities.add_component(id, Activity(activity));
    context.entities.add_component(id, Shootable);
    context.entities.add_component(id, FlungWhenShot);
    context.entities.add_component(id, Life(1));

    let movement_speed = match walk_around_type {
        Lazy => rnd_value(2.0, 4.0),
        Normal => rnd_value(6.0, 12.0),
        Hyperactive => rnd_value(16.0, 20.0),
    };
    context.entities.add_component(id, MovementSpeed(movement_speed));
    context.entities.add_component(id, WalkConstantlyInDirection(walk_direction));

    id
}

pub fn spawn_random_person(context: &mut Core) -> Option<EntityId> {
    if let Some((x, y)) = context.tilemap.get_random_people_spawn_coordinate() {
        let people_sprite = PeopleSprites::rnd();
        let walk_around_type = WalkAroundType::rnd();
        let id = new_person_entity(
            context,
            x * TILE_WIDTH as i32,
            y * TILE_HEIGHT as i32,
            people_sprite,
            walk_around_type,
        );
        Some(id)
    } else {
        None
    }
}

pub fn new_animal_entity(
    context: &mut Core,
    x: i32,
    y: i32,
    animal_sprite: AnimalSprites,
    walk_around_type: WalkAroundType,
) -> EntityId {
    use WalkAroundType::*;

    let id = context.entities.new_entity();

    let activity = EntityActivity::Idle;

    let sprite = Sprite {
        atlas: context.tiles.clone(),
        index: animal_sprite as usize,
        color: animal_sprite.color(),
        flip_horizontally: false,
        flip_vertically: false,
    };

    context.entities.add_component(id, Animal);
    context.entities.add_component(id, Position(Vector2::new(x as f32, y as f32)));
    context.entities.add_component(id, Velocity(Vector2::ZERO));
    context.entities.add_component(id, Forces::new());
    context.entities.add_component(id, Bounds { width: TILE_WIDTH, height: TILE_HEIGHT, radius: TILE_WIDTH / 2});
    context.entities.add_component(id, FacingAngle(0.0));
    context.entities.add_component(id, FlipHorizontallyBasedOnDirection);
    context.entities.add_component(id, sprite);
    context.entities.add_component(id, Activity(activity));
    context.entities.add_component(id, Pusher::new());
    context.entities.add_component(id, Pushable);
    context.entities.add_component(id, Channelable::new(ChannelKind::Throw));
    context.entities.add_component(id, Life(1));
    context.entities.add_component(id, Shootable);

    let (movement_speed, chance_to_move, min_walk_time, max_walk_time, min_walk_cooldown, max_walk_cooldown) = match walk_around_type {
        Lazy => (4.0, 10, 0.5, 2.0, 0.5, 2.0),
        Normal => (8.0, 40, 0.5, 2.0, 0.5, 3.0),
        Hyperactive => (16.0, 90, 0.5, 1.0, 0.5, 2.0),
    };
    context.entities.add_component(id, MovementSpeed(movement_speed));
    context.entities.add_component(id, RandomlyWalksAround::new(min_walk_time, max_walk_time, chance_to_move, min_walk_cooldown, max_walk_cooldown));

    id
}

pub fn spawn_random_animal(context: &mut Core) -> Option<EntityId> {
    if let Some((x, y)) = context.tilemap.get_random_animal_spawn_coordinate() {
        let animal_sprite = AnimalSprites::rnd();
        let walk_around_type = WalkAroundType::rnd();
        let id = new_animal_entity(
            context,
            x * TILE_WIDTH as i32,
            y * TILE_HEIGHT as i32,
            animal_sprite,
            walk_around_type,
        );
        Some(id)
    } else {
        None
    }
}

pub fn new_animation_effect(context: &mut Core, x: i32, y: i32, animation_def: Rc<AnimationDef>, color: u8, delay_scaling_factor: Option<f32>) -> EntityId {
    let id = context.entities.new_entity();

    let sprite = Sprite {
        atlas: context.tiles.clone(),
        index: 0,
        color,
        flip_horizontally: false,
        flip_vertically: false,
    };

    context.entities.add_component(id, Particle);
    context.entities.add_component(id, Position(Vector2::new(x as f32, y as f32)));
    context.entities.add_component(id, sprite);
    context.entities.add_component(id, KillWhenAnimationFinishes);
    context.entities.add_component(id, IgnoresCollision);
    context.entities.add_component(id, IgnoresFriction);
    context.entities.add_component(id, RendersAboveTilemapOverlay);

    let mut animation = AnimationInstance::from(animation_def);
    if let Some(delay_scaling_factor) = delay_scaling_factor {
        animation.delay_override = Some(animation.def.delay * delay_scaling_factor);
    }
    context.entities.add_component(id, animation);

    id
}

pub fn spawn_animation_effect(context: &mut Core, kind: EffectKind, x: i32, y: i32) -> EntityId {
    let animation_def = context.effect_animations.get(&kind).unwrap().clone();
    new_animation_effect(
        context,
        x,
        y,
        animation_def,
        20,
        None
    )
}

pub fn new_pixel_particle(context: &mut Core, x: i32, y: i32, velocity: Vector2, color: u8) -> EntityId {
    let id = context.entities.new_entity();
    context.entities.add_component(id, Particle);
    context.entities.add_component(id, Position(Vector2::new(x as f32, y as f32)));
    context.entities.add_component(id, Velocity(velocity));
    context.entities.add_component(id, Pixel(color));
    context.entities.add_component(id, IgnoresCollision);
    context.entities.add_component(id, IgnoresFriction);
    context.entities.add_component(id, RendersAboveTilemapOverlay);
    id
}

fn add_pixel_particle_animation(context: &mut Core, entity: EntityId, animation_def: Rc<AnimationDef>, delay_scaling_factor: Option<f32>) {
    let mut animation = AnimationInstance::from(animation_def);
    if let Some(delay_scaling_factor) = delay_scaling_factor {
        animation.delay_override = Some(animation.def.delay * delay_scaling_factor);
    }
    context.entities.add_component(entity, KillWhenAnimationFinishes);
    context.entities.add_component(entity, animation);
}

pub fn spawn_pixel_circle(context: &mut Core, x: i32, y: i32, count: usize, speed: f32, lifetime: f32, color: u8) {
    let mut angle = 0.0;
    let randomize = (count as f32).to_radians();
    for _ in 0..count {
        angle += (RADIANS_360 / count as f32).wrapping_radians_add(rnd_value(-randomize, randomize));
        let velocity = Vector2::from_angle(angle) * speed;
        let id = new_pixel_particle(context, x, y, velocity, color);
        context.entities.add_component(id, LifeTime::new(lifetime));
    }
}

pub fn spawn_debris_circle(context: &mut Core, x: i32, y: i32, count: usize, speed: f32) {
    let mut angle = 0.0;
    let randomize = (count as f32).to_radians();
    let animation_def = context.particle_animations.get(&ParticleKind::Debris).unwrap().clone();
    for _ in 0..count {
        angle += (RADIANS_360 / count as f32).wrapping_radians_add(rnd_value(-randomize, randomize));
        let velocity = Vector2::from_angle(angle) * speed;
        let id = new_pixel_particle(context, x, y, velocity, 0);
        add_pixel_particle_animation(context, id, animation_def.clone(), None);
    }
}

pub fn spawn_firey_sparks_circle(context: &mut Core, x: i32, y: i32, count: usize, speed: f32) {
    let mut angle = 0.0;
    let randomize = (count as f32).to_radians();
    let animation_def = context.particle_animations.get(&ParticleKind::FireyDebris).unwrap().clone();
    for _ in 0..count {
        angle += (RADIANS_360 / count as f32).wrapping_radians_add(rnd_value(-randomize, randomize));
        let velocity = Vector2::from_angle(angle) * speed;
        let id = new_pixel_particle(context, x, y, velocity, 0);
        add_pixel_particle_animation(context, id, animation_def.clone(), None);
    }
}

pub fn new_debris_particle(context: &mut Core, x: i32, y: i32, velocity: Vector2, delay_scaling_factor: Option<f32>) -> EntityId {
    let id = new_pixel_particle(context, x, y, velocity, 0);
    let animation_def = context.particle_animations.get(&ParticleKind::Debris).unwrap().clone();
    add_pixel_particle_animation(context, id, animation_def, delay_scaling_factor);
    id
}

pub fn new_firey_debris_particle(context: &mut Core, x: i32, y: i32, velocity: Vector2, delay_scaling_factor: Option<f32>) -> EntityId {
    let id = new_pixel_particle(context, x, y, velocity, 0);
    let animation_def = context.particle_animations.get(&ParticleKind::FireyDebris).unwrap().clone();
    add_pixel_particle_animation(context, id, animation_def, delay_scaling_factor);
    id
}

pub fn new_rising_glitter_particle(context: &mut Core, x: i32, y: i32, velocity: Vector2, delay_scaling_factor: Option<f32>) -> EntityId {
    let id = new_pixel_particle(context, x, y, velocity, 0);
    let animation_def = context.particle_animations.get(&ParticleKind::RisingGlitter).unwrap().clone();
    add_pixel_particle_animation(context, id, animation_def, delay_scaling_factor);
    id
}

pub fn new_rising_magic_particle(context: &mut Core, x: i32, y: i32, velocity: Vector2, delay_scaling_factor: Option<f32>) -> EntityId {
    let id = new_pixel_particle(context, x, y, velocity, 0);
    let animation_def = context.particle_animations.get(&ParticleKind::RisingMagic).unwrap().clone();
    add_pixel_particle_animation(context, id, animation_def, delay_scaling_factor);
    id
}

pub fn spawn_debris_cloud(context: &mut Core, region: Rect) {
    let angle_low = UP - 20.0f32.to_radians();
    let angle_high = UP + 20.0f32.to_radians();
    let speed_low = 7.0;
    let speed_high = 9.0;
    for _ in 0..64 {
        let angle = rnd_value(angle_low, angle_high);
        let speed = rnd_value(speed_low, speed_high);
        let velocity = Vector2::from_angle(angle) * speed;
        new_debris_particle(
            context,
            rnd_value(region.x, region.right()),
            rnd_value(region.y, region.bottom()),
            velocity,
            Some(rnd_value(0.75, 2.0)),
        );
    }
}

pub fn spawn_debris_cloud_at(context: &mut Core, at: EntityId) {
    if let Some(region) = get_entity_rect(&context.entities, at) {
        spawn_debris_cloud(context, region);
    }
}

pub fn spawn_firey_debris_cloud(context: &mut Core, region: Rect) {
    let angle_low = UP - 20.0f32.to_radians();
    let angle_high = UP + 20.0f32.to_radians();
    let speed_low = 7.0;
    let speed_high = 9.0;
    for _ in 0..64 {
        let angle = rnd_value(angle_low, angle_high);
        let speed = rnd_value(speed_low, speed_high);
        let velocity = Vector2::from_angle(angle) * speed;
        new_firey_debris_particle(
            context,
            rnd_value(region.x, region.right()),
            rnd_value(region.y, region.bottom()),
            velocity,
            Some(rnd_value(0.75, 2.0)),
        );
    }
}

pub fn spawn_firey_debris_cloud_at(context: &mut Core, at: EntityId) {
    if let Some(region) = get_entity_rect(&context.entities, at) {
        spawn_firey_debris_cloud(context, region);
    }
}

pub fn spawn_hurt_blood_splatter_effect(context: &mut Core, x: i32, y: i32) {
    // todo
}

pub fn spawn_hurt_blood_splatter_effect_at(context: &mut Core, at: EntityId) {
    if let Some((x, y)) = get_entity_center_xy(&context.entities, at) {
        spawn_hurt_blood_splatter_effect(context, x as i32, y as i32);
    }
}

pub fn spawn_death_blood_splatter_effect(context: &mut Core, x: i32, y: i32) {
    // todo
}

pub fn spawn_death_blood_splatter_effect_at(context: &mut Core, at: EntityId) {
    if let Some((x, y)) = get_entity_center_xy(&context.entities, at) {
        spawn_death_blood_splatter_effect(context, x as i32, y as i32);
    }
}

pub fn spawn_knocked_out_effect(context: &mut Core, x: i32, y: i32) {
    // todo
}

pub fn spawn_knocked_out_effect_at(context: &mut Core, at: EntityId) {
    if let Some((x, y)) = get_entity_center_xy(&context.entities, at) {
        spawn_knocked_out_effect(context, x as i32, y as i32);
    }
}

pub fn spawn_healed_effect(context: &mut Core, region: Rect) {
    // todo
}

pub fn spawn_healed_effect_at(context: &mut Core, at: EntityId) {
    if let Some(region) = get_entity_rect(&context.entities, at) {
        spawn_healed_effect(context, region);
    }
}

pub fn spawn_pickup_effect(context: &mut Core, region: Rect) {
    // todo
}

pub fn spawn_pickup_effect_at(context: &mut Core, at: EntityId, pickup_kind: PickupKind) {
    if let Some(region) = get_entity_rect(&context.entities, at) {
        spawn_pickup_effect(context, region);
    }
}

pub fn new_flung_entity(
    context: &mut Core,
    x: i32,
    y: i32,
    sprite_index: usize,
    color: u8,
    velocity: Vector2,
    distance: f32,
) -> EntityId {
    let id = context.entities.new_entity();

    let sprite = Sprite {
        atlas: context.tiles.clone(),
        index: sprite_index,
        color,
        flip_horizontally: false,
        flip_vertically: false,
    };

    let position = Vector2::new(x as f32, y as f32);

    context.entities.add_component(id, Flung::new(distance));
    context.entities.add_component(id, Position(position));
    context.entities.add_component(id, Velocity(velocity));
    context.entities.add_component(id, Forces::new());
    context.entities.add_component(id, Bounds { width: TILE_WIDTH, height: TILE_HEIGHT, radius: TILE_WIDTH / 2});
    context.entities.add_component(id, sprite);
    context.entities.add_component(id, IgnoresCollision);
    context.entities.add_component(id, IgnoresFriction);
    context.entities.add_component(id, RendersAboveTilemapOverlay);
    context.entities.add_component(id, Transform { angle: 0.0, scale_x: 1.0, scale_y: 1.0 });

    id
}

pub fn new_particle_emitter(
    context: &mut Core,
    center_x: i32,
    center_y: i32,
    radius: u32,
    lifetime: f32,
    kind: ParticleKind,
    min_time: f32,
    max_time: f32,
    min_count: usize,
    max_count: usize,
) -> EntityId {
    let id = context.entities.new_entity();

    let position = Vector2::new(
        (center_x - (radius / 2) as i32) as f32,
        (center_y - (radius / 2) as i32) as f32,
    );
    context.entities.add_component(id, Position(position));
    context.entities.add_component(id, Bounds { width: radius, height: radius, radius });
    context.entities.add_component(id, LifeTime::new(lifetime));
    context.entities.add_component(id, ParticleEmitter::new(kind, min_time, max_time, min_count, max_count));

    id
}

pub fn new_pickupable_entity(context: &mut Core, x: i32, y: i32, force: Option<Force>, kind: PickupKind) -> EntityId {
    let id = context.entities.new_entity();

    let mut forces = Forces::new();
    if let Some(force) = force {
        forces.forces.push(force);
    }

    let sprite = match kind {
        PickupKind::FourLeafClover => MiscSprites::FourLeafClover,
        PickupKind::Heart => MiscSprites::HeartFull,
    };

    context.entities.add_component(id, Pickupable { kind, pre_timer: PICKUP_PRE_TIMER });
    context.entities.add_component(id, Position(Vector2::new(x as f32, y as f32)));
    context.entities.add_component(id, Velocity(Vector2::ZERO));
    context.entities.add_component(id, forces);
    context.entities.add_component(id, Sprite {
        atlas: context.tiles.clone(),
        index: sprite as usize,
        color: sprite.color(),
        flip_horizontally: false,
        flip_vertically: false,
    });
    context.entities.add_component(id, Bounds { width: TILE_WIDTH, height: TILE_HEIGHT, radius: TILE_WIDTH / 2});
    context.entities.add_component(id, LifeTime::new(10.0));
    context.entities.add_component(id, TimedFlicker::new_with_pre_timer(10.0, 7.0, FlickerMethod::OnOff));
    context.entities.add_component(id, Shootable);
    context.entities.add_component(id, Life(1));

    if kind == PickupKind::FourLeafClover {
        context.entities.add_component(id, ParticleEmitter::new(ParticleKind::RisingGlitter, 0.25, 0.5, 1, 3));
    }

    id
}

pub fn maybe_spawn_pickups_from_entity(context: &mut Core, entity: EntityId) {
    let (x, y) = get_entity_xy(&context.entities, entity).unwrap();

    // chance to drop anything
    if rnd_value(0, 99) <= DROP_ANYTHING_CHANCE {
        let kind = match rnd_value(0, 99) {
            0..=10 => {
                context.event_publisher.queue(Event::Tally(TallyEvent::HeartSpawn));
                PickupKind::Heart
            },
            _ => {
                context.event_publisher.queue(Event::Tally(TallyEvent::FourLeafCloverSpawn));
                PickupKind::FourLeafClover
            },
        };

        let angle = (rnd_value(0, 359) as f32).to_radians();
        let force_strength = rnd_value(1.0, 3.0);
        let force = Force {
            force: Vector2::from_angle(angle) * force_strength,
            dissipation_factor: 0.9,
        };
        new_pickupable_entity(context, x as i32, y as i32, Some(force), kind);
    }
}
