use std::path::Path;

use libretrogd::entities::*;
use libretrogd::graphics::*;
use libretrogd::math::*;
use libretrogd::SCREEN_RIGHT;
use libretrogd::states::*;
use sdl2::keyboard::Scancode;

use crate::components::*;
use crate::events::*;
use crate::game::*;
use crate::logic::*;
use crate::spawning::*;
use crate::support::*;
use crate::systems::*;

fn init_systems(context: &mut Game) {
    context.core.entities.remove_all_entities();

    context.core.entities.init_components::<ColorOffset>();
    context.core.entities.init_components::<LifeTime>();
    context.core.entities.init_components::<ColorByLifeTime>();
    context.core.entities.init_components::<TimedFlicker>();
    context.core.entities.init_components::<Transform>();
    context.core.entities.init_components::<RendersAboveTilemapOverlay>();
    context.core.entities.init_components::<WalkingTime>();
    context.core.entities.init_components::<Channelable>();
    context.core.entities.init_components::<Channeler>();
    context.core.entities.init_components::<Position>();
    context.core.entities.init_components::<Bounds>();

    context.component_systems.reset();

    context.component_systems.add_update_system(update_system_world_tick);
    context.component_systems.add_update_system(update_system_moon_tick);
    context.component_systems.add_update_system(update_system_lifetime);
    context.component_systems.add_update_system(update_system_despawn_people_in_despawn_zone);
    context.component_systems.add_update_system(update_system_current_entity_activity);
    context.component_systems.add_update_system(update_system_channeling);
    context.component_systems.add_update_system(update_system_move_channeled_entities);
    context.component_systems.add_update_system(update_system_projectile_hits);
    context.component_systems.add_update_system(update_system_walking_time);
    context.component_systems.add_update_system(update_system_pushing);
    context.component_systems.add_update_system(update_system_movement);
    context.component_systems.add_update_system(update_system_flung);
    context.component_systems.add_update_system(update_system_friction);
    context.component_systems.add_update_system(update_system_force_decay);
    context.component_systems.add_update_system(update_system_randomly_walk_around);
    context.component_systems.add_update_system(update_system_walk_constantly);
    context.component_systems.add_update_system(update_system_animation);
    context.component_systems.add_update_system(update_system_adjust_render_properties_from_animation);
    context.component_systems.add_update_system(update_system_flip_sprite_horizontally_based_on_facing_direction);
    context.component_systems.add_update_system(update_system_timed_flicker);
    context.component_systems.add_update_system(update_system_transforms);
    context.component_systems.add_update_system(update_system_pickups);
    context.component_systems.add_update_system(update_system_check_target_positions);
    context.component_systems.add_update_system(update_system_particle_emitter);
    context.component_systems.add_update_system(update_system_life_and_death);

    context.component_systems.add_render_system(render_system_world_base);
    context.component_systems.add_render_system(render_system_render_sprites);
    context.component_systems.add_render_system(render_system_render_pixels);
    context.component_systems.add_render_system(render_system_world_overlay);
    context.component_systems.add_render_system(render_system_render_sprites_top);
    context.component_systems.add_render_system(render_system_render_pixels_top);
    context.component_systems.add_render_system(render_system_effects);

    context.event_listeners.clear();
    context.event_listeners.add(event_listener);
    context.core.event_publisher.clear();
}

///////////////////////////////////////////////////////////////////////////////////////////////////

pub struct MainMenuState {
    fade: f32,
    full_transition: bool,
    selection: i32,
}

impl MainMenuState {
    pub fn new() -> Self {
        MainMenuState {
            fade: 0.0,
            full_transition: true,
            selection: 0,
        }
    }

    fn init(&mut self, context: &mut Game) {
        init_systems(context);
        context.core.tilemap = context.core.title_screen_tilemap.clone();
        new_world_entity(&mut context.core, 8, 8, None);
        new_moon_entity(&mut context.core, 24, 20);
    }
}

impl GameState<Game> for MainMenuState {
    fn update(&mut self, state: State, context: &mut Game) -> Option<StateChange<Game>> {
        if state == State::Active {
            if context.core.system.keyboard.is_key_pressed(Scancode::Escape) {
                play_sound_menu_select(&mut context.core);
                return Some(StateChange::Pop(1));
            }
            if context.core.system.keyboard.is_key_pressed(Scancode::Up) {
                play_sound_menu_move(&mut context.core);
                self.selection = (self.selection - 1).clamp(0, 3);
            }
            if context.core.system.keyboard.is_key_pressed(Scancode::Down) {
                play_sound_menu_move(&mut context.core);
                self.selection = (self.selection + 1).clamp(0, 3);
            }
            if context.core.system.keyboard.is_key_pressed(Scancode::Return) {
                play_sound_menu_select(&mut context.core);
                match self.selection {
                    0 => {
                        self.full_transition = true;
                        return Some(StateChange::Push(Box::new(GamePlayState::new())))
                    },
                    1 => {
                        self.full_transition = false;
                        return Some(StateChange::Push(Box::new(InstructionsState::new())))
                    },
                    2 => {
                        self.full_transition = false;
                        return Some(StateChange::Push(Box::new(AboutState)))
                    },
                    3 => {
                        self.full_transition = true;
                        return Some(StateChange::Pop(1))
                    },
                    _ => {},
                }
            }
        }

        context.do_events();
        context.component_systems.update(&mut context.core);

        None
    }

    fn render(&mut self, state: State, context: &mut Game) {
        context.core.system.video.clear(0);
        context.component_systems.render(&mut context.core);

        print_string_with_shadow(&mut context.core.system.video, "Lucky Moon\n Rock Bridge", 70, 8, 19, &context.core.title_font);

        let y = 32;
        let x = 32;
        let width = 88;
        let height = 48;
        const SPACER: i32 = 8;
        draw_window(&mut context.core.system.video, &context.core.tiles, x, y, x + width, y + height);

        let x = x + SPACER;
        let y = y + SPACER + 1;

        let selection_y = y + (self.selection as i32 * 8);
        print_string(&mut context.core, ">", x + 2, selection_y, 21);

        let x = x + SPACER;
        print_string(&mut context.core, "Begin", x, y + (0 * SPACER), 21);
        print_string(&mut context.core, "How To Play", x, y + (1 * SPACER), 21);
        print_string(&mut context.core, "About", x, y + (2 * SPACER), 21);
        print_string(&mut context.core, "Quit", x, y + (3 * SPACER), 21);
    }

    fn transition(&mut self, state: State, context: &mut Game) -> bool {
        if self.full_transition {
            update_fade_transition(state, &mut self.fade, context.core.delta * 3.0, context)
        } else {
            true
        }
    }

    fn state_change(&mut self, new_state: State, old_state: State, context: &mut Game) {
        match new_state {
            State::Pending | State::Resume => {
                if self.full_transition {
                    self.init(context)
                }
            },
            State::TransitionIn => {
                self.fade = 0.0;
            },
            State::TransitionOut(_) => {
                self.fade = 1.0;
            },
            State::Paused => {
                context.core.system.palette = context.core.palette.clone();
            }
            _ => {}
        }
    }
}

///////////////////////////////////////////////////////////////////////////////////////////////////

pub struct InstructionsState {
    pub page: i32,
}

impl InstructionsState {
    pub fn new() -> Self {
        InstructionsState {
            page: 0,
        }
    }
}

impl GameState<Game> for InstructionsState {
    fn update(&mut self, state: State, context: &mut Game) -> Option<StateChange<Game>> {
        if state == State::Active {
            if context.core.system.keyboard.is_key_pressed(Scancode::Escape) {
                play_sound_menu_select(&mut context.core);
                return Some(StateChange::Pop(1));
            }
            if context.core.system.keyboard.is_key_pressed(Scancode::Up) ||
                context.core.system.keyboard.is_key_pressed(Scancode::Left) {
                play_sound_menu_move(&mut context.core);
                self.page = (self.page - 1).clamp(0, 5);
            }
            if context.core.system.keyboard.is_key_pressed(Scancode::Down) ||
                context.core.system.keyboard.is_key_pressed(Scancode::Right) {
                play_sound_menu_move(&mut context.core);
                self.page = (self.page + 1).clamp(0, 5);
            }
            if context.core.system.keyboard.is_key_pressed(Scancode::Return) ||
                context.core.system.keyboard.is_key_pressed(Scancode::Space) {
                if self.page == 5 {
                    play_sound_menu_select(&mut context.core);
                    return Some(StateChange::Pop(1));
                } else {
                    play_sound_menu_move(&mut context.core);
                    self.page = (self.page + 1).clamp(0, 5);
                }
            }
        }

        context.do_events();
        context.component_systems.update(&mut context.core);

        None
    }

    fn render(&mut self, state: State, context: &mut Game) {
        context.core.system.video.clear(0);
        context.component_systems.render(&mut context.core);

        draw_window(&mut context.core.system.video, &context.core.tiles, 8, 8, 152, 112);

        let y = 17;
        print_title_string(&mut context.core, "How To Play", 44, y, 9);
        let y = y + 12;

        match self.page {
            0 => {
                print_string(&mut context.core, "You play a wizard!", 16, y, 20);
                blit_tile(&mut context.core, PLAYER_SPRITE as usize, 110, y - 1, PLAYER_SPRITE_COLOR);

                let y = y + 12;
                print_string(&mut context.core, "Your goal is to get as\nmany four leaf clovers\nas you can.", 16, y, 20);
                blit_tile(&mut context.core, MiscSprites::FourLeafClover as usize, 130, y + 7, MiscSprites::FourLeafClover.color());

                let y = y + 26;
                print_string(&mut context.core, "You know how to conjure\nfour leaf clovers from\npeople recently struck\nby moon rocks.", 16, y, 20);
                blit_tile(&mut context.core, MiscSprites::MoonRock as usize, 100, y + 26, MiscSprites::MoonRock.color());
                blit_tile(&mut context.core, PeopleSprites::StrongVillager as usize, 118, y + 26, PeopleSprites::StrongVillager.color());
            },
            1 => {
                print_string(&mut context.core, "You will need to get\nthe moon to throw\nmoon rocks at you!", 16, y, 20);
                blit_tile(&mut context.core, MiscSprites::Moon as usize, 120, y + 7, MiscSprites::Moon.color());

                let y = y + 26;
                print_string(&mut context.core, "Throw animals at the moon\nto make it angry at you.", 16, y, 20);

                let y = y + 20;
                blit_tile(&mut context.core, PLAYER_SPRITE as usize, 32, y + 8, 68);
                context.core.system.video.circle(32+4, y+8+4, 12, 19);
                blit_tile(&mut context.core, AnimalSprites::Pig as usize, 48, y + 8, AnimalSprites::Pig.color());

                context.core.system.video.blit_region(
                    BlitMethod::RotoZoomTransparentOffset {
                        transparent_color: 0,
                        offset: AnimalSprites::Goose.color(),
                        angle: 89.0f32.to_radians(),
                        scale_x: 1.5,
                        scale_y: 1.5,
                    },
                    context.core.tiles.bitmap(),
                    &context.core.tiles[AnimalSprites::Goose as usize],
                    70, y + 3
                );
                context.core.system.video.blit_region(
                    BlitMethod::RotoZoomTransparentOffset {
                        transparent_color: 0,
                        offset: AnimalSprites::Dog.color(),
                        angle: 154.0f32.to_radians(),
                        scale_x: 1.2,
                        scale_y: 1.2,
                    },
                    context.core.tiles.bitmap(),
                    &context.core.tiles[AnimalSprites::Dog as usize],
                    90, y + 3
                );
                blit_tile(&mut context.core, MiscSprites::Moon as usize, 110, y + 8, 246);
            },
            2 => {
                print_string(&mut context.core, "When angry, the moon will\nshoot at you. Stand near\nthe people on the bridge\nand be waiting to absorb\nanyone flung up in the air\nwhen they get hit!", 16, y, 20);

                let y = y + 46;
                blit_tile(&mut context.core, PLAYER_SPRITE as usize, 90, y + 8, 100);
                context.core.system.video.circle(90+4, y+8+4, 12, 8);
                context.core.system.video.blit_region(
                    BlitMethod::RotoZoomTransparentOffset {
                        transparent_color: 0,
                        offset: PeopleSprites::Guard.color(),
                        angle: 39.0f32.to_radians(),
                        scale_x: 1.5,
                        scale_y: 1.5,
                    },
                    context.core.tiles.bitmap(),
                    &context.core.tiles[PeopleSprites::Guard as usize],
                    39, y + 3
                );
                context.core.system.video.blit_region(
                    BlitMethod::RotoZoomTransparentOffset {
                        transparent_color: 0,
                        offset: PeopleSprites::Ninja.color(),
                        angle: 211.0f32.to_radians(),
                        scale_x: 1.2,
                        scale_y: 1.2,
                    },
                    context.core.tiles.bitmap(),
                    &context.core.tiles[PeopleSprites::Ninja as usize],
                    60, y + 3
                );
                context.core.system.video.blit_region(
                    BlitMethod::RotoZoomTransparentOffset {
                        transparent_color: 0,
                        offset: PeopleSprites::VillagerA.color(),
                        angle: 300.0f32.to_radians(),
                        scale_x: 1.6,
                        scale_y: 1.6,
                    },
                    context.core.tiles.bitmap(),
                    &context.core.tiles[PeopleSprites::VillagerA as usize],
                    75, y + 7
                );
            },
            3 => {
                print_string(&mut context.core, "People that are absorbed\nsometimes turn into four\nleaf clovers.", 16, y, 20);

                let y = y + 26;
                print_string(&mut context.core, "Others will turn into\nhearts!", 16, y, 20);
                blit_tile(&mut context.core, MiscSprites::HeartFull as usize, 125, y + 4, MiscSprites::HeartFull.color());

                let y = y + 20;
                print_string(&mut context.core, "You are not invincible! Be\nsure to pick up hearts if\nyou've been hit by too\nmany moon rocks!", 16, y, 20);
            },
            4 => {
                print_title_string(&mut context.core, "Tip!", 70, y, 8);

                let y = y + 12;
                print_string(&mut context.core, "The more angry the moon\nis at you, the more often\nit will shoot at you.", 16, y, 20);

                let y = y + 26;
                print_string(&mut context.core, "The moon's colour\nindicates how angry it is\nat you.", 16, y, 20);

                let y = y + 22;
                blit_tile(&mut context.core, MiscSprites::Moon as usize, 60, y + 4, 240);
                blit_tile(&mut context.core, MiscSprites::Moon as usize, 76, y + 4, 243);
                blit_tile(&mut context.core, MiscSprites::Moon as usize, 92, y + 4, 246);

            },
            _ => {
                print_title_string(&mut context.core, "Controls", 52, y, 8);

                let y = y + 12;
                print_string(&mut context.core, "Use the arrow keys to\nmove around.", 16, y, 20);

                let y = y + 20;
                print_string(&mut context.core, "Z key to\nchannel absorb", 16, y, 20);
                blit_tile(&mut context.core, PLAYER_SPRITE as usize, 110, y + 3, 100);
                context.core.system.video.circle(110+4, y+3+4, 10, 8);

                let y = y + 25;
                print_string(&mut context.core, "X key to\nchannel throw", 16, y, 20);
                blit_tile(&mut context.core, PLAYER_SPRITE as usize, 110, y + 3, 68);
                context.core.system.video.circle(110+4, y+3+4, 10, 19);
            },
        }
    }

    fn transition(&mut self, state: State, context: &mut Game) -> bool {
        true
    }

    fn state_change(&mut self, new_state: State, old_state: State, context: &mut Game) {
    }
}

///////////////////////////////////////////////////////////////////////////////////////////////////

pub struct AboutState;

impl GameState<Game> for AboutState {
    fn update(&mut self, state: State, context: &mut Game) -> Option<StateChange<Game>> {
        if state == State::Active {
            if context.core.system.keyboard.is_key_pressed(Scancode::Escape) ||
                context.core.system.keyboard.is_key_pressed(Scancode::Return) ||
                context.core.system.keyboard.is_key_pressed(Scancode::Space) {
                play_sound_menu_select(&mut context.core);
                return Some(StateChange::Pop(1));
            }
        }

        context.do_events();
        context.component_systems.update(&mut context.core);

        None
    }

    fn render(&mut self, state: State, context: &mut Game) {
        context.core.system.video.clear(0);
        context.component_systems.render(&mut context.core);

        draw_window(&mut context.core.system.video, &context.core.tiles, 8, 8, 152, 112);

        context.core.system.video.blit_region(BlitMethod::Solid, &context.core.emojis.bitmap(), &context.core.emojis[0], 40, 20);
        context.core.system.video.blit_region(BlitMethod::Solid, &context.core.emojis.bitmap(), &context.core.emojis[1], 40+19, 20);
        context.core.system.video.blit_region(BlitMethod::Solid, &context.core.emojis.bitmap(), &context.core.emojis[2], 80, 20);
        context.core.system.video.blit_region(BlitMethod::Solid, &context.core.emojis.bitmap(), &context.core.emojis[3], 80+19, 20);
        print_string(&mut context.core, "A game produced during\nthe GDR 4x4x4 Challenge\nfor June 2022\n", 20, 40, 21);
        print_string(&mut context.core, "http://noop.rocks/gdr/", 24, 64, 16);
        context.core.system.video.horiz_line(32, 120, 74, 1);
        print_string(&mut context.core, "By: Gered King", 35, 78, 22);
        print_string(&mut context.core, "GFX: @dennisbusch_de", 28, 88, 22);
    }

    fn transition(&mut self, state: State, context: &mut Game) -> bool {
        true
    }

    fn state_change(&mut self, new_state: State, old_state: State, context: &mut Game) {
    }
}

///////////////////////////////////////////////////////////////////////////////////////////////////

pub struct GamePlayState {
    fade: f32,
    in_menu: bool,
    menu_selection: i32,
    full_transition: bool,
}

impl GamePlayState {
    pub fn new() -> Self {
        GamePlayState {
            fade: 0.0,
            in_menu: false,
            menu_selection: 0,
            full_transition: true,
        }
    }

    fn init(&mut self, context: &mut Game) {
        init_systems(context);

        context.core.tilemap = context.core.game_tilemap.clone();

        new_world_entity(&mut context.core, 8, 8, Some(GAME_TIME));
        new_player_entity(&mut context.core, 64, 96);
        new_moon_entity(&mut context.core, MOON_X, MOON_Y);
    }
}

impl GameState<Game> for GamePlayState {
    fn update(&mut self, state: State, context: &mut Game) -> Option<StateChange<Game>> {
        if state == State::Active {
            if self.in_menu {
                if context.core.system.keyboard.is_key_pressed(Scancode::Escape) {
                    self.in_menu = false;
                    play_sound_menu_select(&mut context.core);
                }
                if context.core.system.keyboard.is_key_pressed(Scancode::Up) {
                    self.menu_selection = (self.menu_selection - 1).clamp(0, 1);
                    play_sound_menu_move(&mut context.core);
                }
                if context.core.system.keyboard.is_key_pressed(Scancode::Down) {
                    self.menu_selection = (self.menu_selection + 1).clamp(0, 1);
                    play_sound_menu_move(&mut context.core);
                }
                if context.core.system.keyboard.is_key_pressed(Scancode::Return) {
                    play_sound_menu_select(&mut context.core);
                    match self.menu_selection {
                        0 => self.in_menu = false,
                        _ => return Some(StateChange::Pop(1)),
                    };
                }
            } else {
                if context.core.system.keyboard.is_key_pressed(Scancode::Escape) {
                    self.in_menu = true;
                    self.menu_selection = 0;
                    play_sound_menu_select(&mut context.core);
                }

                if let Some((player, _)) = context.core.entities.components::<Player>().single() {
                    if context.core.system.keyboard.is_key_down(Scancode::Up) {
                        context.core.event_publisher.queue(Event::TurnAndMove(*player, RADIANS_270));
                    }
                    if context.core.system.keyboard.is_key_down(Scancode::Down) {
                        context.core.event_publisher.queue(Event::TurnAndMove(*player, RADIANS_90));
                    }
                    if context.core.system.keyboard.is_key_down(Scancode::Left) {
                        context.core.event_publisher.queue(Event::TurnAndMove(*player, RADIANS_180));
                    }
                    if context.core.system.keyboard.is_key_down(Scancode::Right) {
                        context.core.event_publisher.queue(Event::TurnAndMove(*player, RADIANS_0));
                    }
                    if context.core.system.keyboard.is_key_down(Scancode::Z) {
                        context.core.event_publisher.queue(Event::Channel { entity: *player, kind: ChannelKind::Abosrb });
                    }
                    if context.core.system.keyboard.is_key_down(Scancode::X) {
                        context.core.event_publisher.queue(Event::Channel { entity: *player, kind: ChannelKind::Throw });
                    }
                }

                context.do_events();
                context.component_systems.update(&mut context.core);

                if let Some((_, world)) = context.core.entities.components::<World>().single() {
                    let times_up = world.remaining_time == Some(0.0);
                    let died = world.time_to_game_over == Some(0.0);
                    if times_up || died {
                        self.full_transition = false;
                        return Some(StateChange::Swap(Box::new(GameOverState::new(died))));
                    }
                }
            }
        }

        None
    }

    fn render(&mut self, state: State, context: &mut Game) {
        context.core.system.video.clear(0);
        context.component_systems.render(&mut context.core);
        render_status_bar(&mut context.core);

        if self.in_menu {
            let y = 32;
            let x = 32;
            let width = 88;
            let height = 48;
            const SPACER: i32 = 8;
            draw_window(&mut context.core.system.video, &context.core.tiles, x, y, x + width, y + height);


            let x = x + SPACER;
            let y = y + SPACER + 1;
            print_title_string(&mut context.core, "Paused!", x + 14, y, 8);

            let y = y + 16;
            let selection_y = y + (self.menu_selection as i32 * 8);
            print_string(&mut context.core, ">", x + 2, selection_y, 21);

            let x = x + SPACER;
            print_string(&mut context.core, "Continue", x, y + (0 * SPACER), 21);
            print_string(&mut context.core, "Quit To Menu", x, y + (1 * SPACER), 21);
        }
    }

    fn transition(&mut self, state: State, context: &mut Game) -> bool {
        if self.full_transition {
            update_fade_transition(state, &mut self.fade, context.core.delta * 3.0, context)
        } else {
            true
        }
    }

    fn state_change(&mut self, new_state: State, old_state: State, context: &mut Game) {
        match new_state {
            State::Pending | State::Resume =>
            if self.full_transition {
                self.init(context);
            },
            State::TransitionIn => {
                self.fade = 0.0;
            },
            State::TransitionOut(_) => {
                self.fade = 1.0;
            },
            State::Paused => {
                context.core.system.palette = context.core.palette.clone();
            },
            _ => {}
        }
    }
}

fn render_status_bar_heart(core: &mut Core, x: i32, value: i32) {
    let sprite = match value {
        2 => MiscSprites::HeartFull,
        1 => MiscSprites::HeartHalf,
        _ => MiscSprites::HeartEmpty,
    };
    blit_tile(core, sprite as usize, x, 0, sprite.color());
}

fn render_status_bar(core: &mut Core) {
    let num_clovers;
    let current_life;
    let minutes;
    let seconds;
    {
        let worlds = core.entities.components::<World>();
        let tallys = core.entities.components::<Tally>();
        let players = core.entities.components::<Player>();

        let (world_id, world) = worlds.single().unwrap();

        let total_seconds = world.remaining_time.unwrap() as u32;
        minutes = total_seconds / 60;
        seconds = total_seconds % 60;

        if let Some(tally) = tallys.get(world_id) {
            num_clovers = *tally.0.get(&TallyEvent::FourLeafCloverPickup).unwrap_or(&0);
        } else {
            num_clovers = 0;
        }

        if let Some((player_id, _)) = players.single() {
            let lifes = core.entities.components::<Life>();
            let life = lifes.get(player_id).unwrap();
            current_life = life.0;
        } else {
            current_life = 0;
        }
    }

    const CLOCK_POS: i32 = 56;
    const CLOVER_POS: i32 = 0;
    const LIFE_POS: i32 = 136;

    core.system.video.filled_rect(0, 0, SCREEN_RIGHT as i32, 7, 0);
    core.system.video.horiz_line(0, SCREEN_RIGHT as i32, 8, 1);

    blit_tile(core, MiscSprites::FourLeafClover as usize, CLOVER_POS, 0, MiscSprites::FourLeafClover.color());
    print_string(core, &format!("{}", num_clovers), CLOVER_POS + 9, 1, 21);

    blit_tile(core, MiscSprites::Clock as usize, CLOCK_POS, 0, MiscSprites::Clock.color());
    print_string(core, &format!("{:02}:{:02}", minutes, seconds), CLOCK_POS + 9, 1, 21);

    let mut remaining_life_count = current_life;
    for i in 0..3 {
        let x = LIFE_POS + (i * 8);
        let value = if remaining_life_count >= 2 {
            2
        } else {
            remaining_life_count
        };
        render_status_bar_heart(core, x, value);
        remaining_life_count -= 2;
    }
}

pub struct GameOverState {
    pub died: bool,
    pub fade: f32,
}

impl GameOverState {
    pub fn new(died: bool) -> Self {
        GameOverState {
            died,
            fade: 1.0,
        }
    }
    pub fn init(&mut self, context: &mut Game) {

    }
}

impl GameState<Game> for GameOverState {
    fn update(&mut self, state: State, context: &mut Game) -> Option<StateChange<Game>> {
        if state == State::Active {
            if context.core.system.keyboard.is_key_pressed(Scancode::Return) ||
                context.core.system.keyboard.is_key_pressed(Scancode::Escape) ||
                context.core.system.keyboard.is_key_pressed(Scancode::Space) {
                play_sound_menu_select(&mut context.core);
                return Some(StateChange::Pop(1));
            }
        }

        None
    }

    fn render(&mut self, state: State, context: &mut Game) {
        context.core.system.video.clear(0);
        context.component_systems.render(&mut context.core);
        render_status_bar(&mut context.core);

        draw_window(&mut context.core.system.video, &context.core.tiles, 8, 8, 152, 112);
        if self.died {
            print_title_string(&mut context.core, "You died!", 50, 20, 28);
        } else {
            print_title_string(&mut context.core, "Time's up!", 48, 20, 8);
        }

        let world_entity_id = get_world_entity_id(&context.core.entities).unwrap();
        let tallys = context.core.entities.components::<Tally>();
        let tally = tallys.get(&world_entity_id).unwrap();

        let people_spawned = *tally.0.get(&TallyEvent::PeopleSpawn).unwrap_or(&0);
        let people_flung = *tally.0.get(&TallyEvent::PeopleFlung).unwrap_or(&0);
        let people_killed = *tally.0.get(&TallyEvent::PeopleKill).unwrap_or(&0);
        let people_absorbed = *tally.0.get(&TallyEvent::PeopleAbsorb).unwrap_or(&0);

        let animal_spawned = *tally.0.get(&TallyEvent::AnimalSpawn).unwrap_or(&0);
        let animal_flung = *tally.0.get(&TallyEvent::AnimalThrow).unwrap_or(&0);
        let animal_killed = *tally.0.get(&TallyEvent::AnimalKill).unwrap_or(&0);

        let moon_rock_spawned = *tally.0.get(&TallyEvent::MoonRockSpawn).unwrap_or(&0);
        let moon_rock_hit_player = *tally.0.get(&TallyEvent::MoonRockPlayerHit).unwrap_or(&0);

        let clover_spawned = *tally.0.get(&TallyEvent::FourLeafCloverSpawn).unwrap_or(&0);
        let clover_picked_up = *tally.0.get(&TallyEvent::FourLeafCloverPickup).unwrap_or(&0);

        let heart_spawned = *tally.0.get(&TallyEvent::HeartSpawn).unwrap_or(&0);
        let heart_picked_up = *tally.0.get(&TallyEvent::HeartPickup).unwrap_or(&0);

        drop(tallys);

        // ---------------------------------------------------------------------------------------


        const COLUMN_1_X: i32 = 72;
        const COLUMN_2_X: i32 = 96;
        const COLUMN_3_X: i32 = 120;

        // ---------------------------------------------------------------------------------------

        let y = 32;
        blit_tile(&mut context.core, PeopleSprites::VillagerA as usize, COLUMN_1_X + 2, y, PeopleSprites::VillagerA.color());
        blit_tile(&mut context.core, AnimalSprites::Goose as usize, COLUMN_2_X + 2, y, AnimalSprites::Goose.color());
        blit_tile(&mut context.core, MiscSprites::MoonRock as usize, COLUMN_3_X + 2, y, MiscSprites::MoonRock.color());

        let y = y + 9;
        print_string(&mut context.core, "Total\nFlung\nKilled\nAbsorbed", 16, y, 21);
        print_string(&mut context.core, &format!("{}\n{}\n{}\n{}", people_spawned, people_flung, people_killed, people_absorbed), COLUMN_1_X, y, 20);
        print_string(&mut context.core, &format!("{}\n{}\n{}\n{}", animal_spawned, animal_flung, animal_killed, "--"), COLUMN_2_X, y, 20);
        print_string(&mut context.core, &format!("{}\n{}\n{}\n{}", moon_rock_spawned, "--", "--", "--"), COLUMN_3_X, y, 20);

        // ---------------------------------------------------------------------------------------

        let y = 71;
        blit_tile(&mut context.core, MiscSprites::FourLeafClover as usize, COLUMN_1_X + 2, y, MiscSprites::FourLeafClover.color());
        blit_tile(&mut context.core, MiscSprites::HeartFull as usize, COLUMN_2_X + 2, y, MiscSprites::HeartFull.color());

        let y = y + 9;
        print_string(&mut context.core, "Total\nPicked up", 16, y, 21);
        print_string(&mut context.core, &format!("{}\n{}", clover_spawned, clover_picked_up), COLUMN_1_X, y, 20);
        print_string(&mut context.core, &format!("{}\n{}", heart_spawned, heart_picked_up), COLUMN_2_X, y, 20);

        // ---------------------------------------------------------------------------------------

        let y = 98;
        print_string(&mut context.core, "Total Damage Taken:", 16, y, 21);
        print_string(&mut context.core, &format!("{}", moon_rock_hit_player as u32 * MOON_ROCK_DAMAGE), 120, y, 20);
    }

    fn transition(&mut self, state: State, context: &mut Game) -> bool {
        update_fade_transition(state, &mut self.fade, context.core.delta * 3.0, context)
    }

    fn state_change(&mut self, new_state: State, old_state: State, context: &mut Game) {
        match new_state {
            State::Pending | State::Resume => {
                self.init(context);
            },
            State::TransitionIn => {},
            State::TransitionOut(_) => {
                self.fade = 1.0;
            },
            State::Paused => {
                context.core.system.palette = context.core.palette.clone();
            },
            _ => {}
        }
    }
}
