use std::fs::File;
use std::io::{BufReader, Seek};
use std::path::Path;
use std::rc::Rc;

use anyhow::{Context, Result};
use byteorder::ReadBytesExt;
use libretrogd::audio::*;
use libretrogd::graphics::*;
use libretrogd::states::*;
use libretrogd::utils::rnd_value;

use crate::Game;
use crate::game::Core;
use crate::tilemap::{TILE_HEIGHT, TILE_WIDTH, TileMap};

fn open_file_as_byte_reader(path: &Path) -> Result<BufReader<File>> {
    let f = File::open(path).context(format!("Opening file: {:?}", path))?;
    Ok(BufReader::new(f))
}

pub fn load_palette(path: &Path) -> Result<Palette> {
    let mut reader = open_file_as_byte_reader(path).context(format!("Loading palette file: {:?}", path))?;
    load_palette_bytes(&mut reader, path.to_str().unwrap())
}

pub fn load_palette_bytes<T: ReadBytesExt>(reader: &mut T, name: &str) -> Result<Palette> {
    Palette::load_from_bytes(reader, PaletteFormat::Vga).context(format!("Loading palette: {:?}", name))
}

pub fn load_font(path: &Path) -> Result<BitmaskFont> {
    let mut reader = open_file_as_byte_reader(path).context(format!("Loading font file: {:?}", path))?;
    load_font_bytes(&mut reader, path.to_str().unwrap())
}

pub fn load_font_bytes<T: ReadBytesExt>(reader: &mut T, name: &str) -> Result<BitmaskFont> {
    BitmaskFont::load_from_bytes(reader).context(format!("Loading font: {:?}", name))
}

pub fn load_bitmap_atlas_autogrid(path: &Path, tile_width: u32, tile_height: u32) -> Result<BitmapAtlas> {
    let mut reader = open_file_as_byte_reader(path).context(format!("Loading bitmap file: {:?}", path))?;
    load_bitmap_bytes_atlas_autogrid(&mut reader, path.to_str().unwrap(), tile_width, tile_height)
}

pub fn load_bitmap_bytes_atlas_autogrid<T: ReadBytesExt + Seek>(reader: &mut T, name: &str, tile_width: u32, tile_height: u32) -> Result<BitmapAtlas> {
    let (bmp, _) = Bitmap::load_pcx_bytes(reader).context(format!("Loading bitmap atlas: {:?}", name))?;
    let mut atlas = BitmapAtlas::new(bmp);
    atlas.add_grid(tile_width, tile_height)?;
    Ok(atlas)
}

pub fn load_bitmap_atlas(path: &Path) -> Result<BitmapAtlas> {
    let mut reader = open_file_as_byte_reader(path).context(format!("Loading bitmap file: {:?}", path))?;
    load_bitmap_bytes_atlas(&mut reader, path.to_str().unwrap())
}

pub fn load_bitmap_bytes_atlas<T: ReadBytesExt + Seek>(reader: &mut T, name: &str) -> Result<BitmapAtlas> {
    let (bmp, _) = Bitmap::load_pcx_bytes(reader).context(format!("Loading bitmap atlas: {:?}", name))?;
    let atlas = BitmapAtlas::new(bmp);
    Ok(atlas)
}

pub fn load_tilemap(path: &Path) -> Result<TileMap> {
    TileMap::load_file(path).context(format!("Loading tilemap: {:?}", path))
}

pub fn load_tilemap_bytes<T: ReadBytesExt>(reader: &mut T, name: &str) -> Result<TileMap> {
    TileMap::load_bytes(reader).context(format!("Loading tilemap: {:?}", name))
}

pub fn load_wav(path: &Path, desired_spec: &AudioSpec) -> Result<AudioBuffer> {
    let mut reader = open_file_as_byte_reader(path).context(format!("Loading wav file: {:?}", path))?;
    load_wav_bytes(&mut reader, path.to_str().unwrap(), desired_spec)
}

pub fn load_wav_bytes<T: ReadBytesExt + Seek>(reader: &mut T, name: &str, desired_spec: &AudioSpec) -> Result<AudioBuffer> {
    let mut audio_buffer = AudioBuffer::load_wav_bytes(reader).context(format!("Loading wav: {:?}", name))?;
    audio_buffer.convert(desired_spec).context(format!("Converting wav {:?} to spec {:?}", name, desired_spec))
}

pub fn draw_window(dest: &mut Bitmap, tiles: &BitmapAtlas, left: i32, top: i32, right: i32, bottom: i32) {
    const TILE_SIZE: i32 = 8;

    let blit_opts = BlitMethod::TransparentOffset { transparent_color: 0, offset: 48 };

    dest.filled_rect(left+6, top+6, right-6, bottom-6, 0);

    // corners
    dest.blit_region(blit_opts, &tiles.bitmap(), &tiles[285], left + 1, top + 1);
    dest.blit_region(blit_opts, &tiles.bitmap(), &tiles[273], right - TILE_SIZE, top + 1);
    dest.blit_region(blit_opts, &tiles.bitmap(), &tiles[287], left + 1, bottom - TILE_SIZE);
    dest.blit_region(blit_opts, &tiles.bitmap(), &tiles[301], right - TILE_SIZE, bottom - TILE_SIZE);

    // top and bottom edges
    for i in 0..((right - left) / TILE_SIZE) - 2 {
        let x = left + TILE_SIZE + (i * TILE_SIZE);
        dest.blit_region(blit_opts, &tiles.bitmap(), &tiles[271], x, top + 1);
        dest.blit_region(blit_opts, &tiles.bitmap(), &tiles[299], x, bottom - TILE_SIZE);
    }

    // left and right edges
    for i in 0..((bottom - top) / TILE_SIZE) - 2 {
        let y = top + TILE_SIZE + (i * TILE_SIZE);
        dest.blit_region(blit_opts, &tiles.bitmap(), &tiles[284], left + 1, y);
        dest.blit_region(blit_opts, &tiles.bitmap(), &tiles[286], right - TILE_SIZE, y);
    }
}

pub fn update_fade_transition(state: State, fade: &mut f32, delta: f32, context: &mut Game) -> bool {
    match state {
        State::TransitionIn => {
            *fade += delta;
            if *fade >= 1.0 {
                *fade = 1.0;
                context.core.system.palette = context.core.palette.clone();
                true
            } else {
                context.core.system.palette.lerp(0..=255, &context.core.fade_out_palette, &context.core.palette, *fade);
                false
            }
        },
        State::TransitionOut(_) => {
            *fade -= delta;
            if *fade <= 0.0 {
                *fade = 0.0;
                context.core.system.palette = context.core.fade_out_palette.clone();
                true
            } else {
                context.core.system.palette.lerp(0..=255, &context.core.fade_out_palette, &context.core.palette, *fade);
                false
            }
        },
        _ => {
            true
        }
    }
}

pub fn print_string(core: &mut Core, text: &str, x: i32, y: i32, color: u8) {
    core.system.video.print_string(text, x, y, FontRenderOpts::Color(color), &core.font);
}

pub fn print_title_string(core: &mut Core, text: &str, x: i32, y: i32, color: u8) {
    core.system.video.print_string(text, x, y, FontRenderOpts::Color(color), &core.title_font);
}

pub fn blit_tile(core: &mut Core, tile_index: usize, x: i32, y: i32, color: u8) {
    core.system.video.blit_region(
        BlitMethod::TransparentOffset {
            transparent_color: 0,
            offset: color,
        },
        core.tiles.bitmap(),
        &core.tiles[tile_index],
        x, y
    );
}

pub fn print_string_with_shadow(bitmap: &mut Bitmap, text: &str, x: i32, y: i32, color: u8, font: &BitmaskFont) {
    bitmap.print_string(text, x + 1, y + 1, FontRenderOpts::Color(1), font);
    bitmap.print_string(text, x, y, FontRenderOpts::Color(color), font);
}

pub fn play_looping_sound(core: &mut Core, buffer: Rc<AudioBuffer>) {
    core.system.audio_queue.play_buffer_rc(buffer, true).unwrap();
}

pub fn play_sound(core: &mut Core, buffer: Rc<AudioBuffer>) {
    core.system.audio_queue.play_buffer_rc(buffer, false).unwrap();
}

fn get_random_buffer(buffers: &Vec<Rc<AudioBuffer>>) -> Rc<AudioBuffer> {
    let index = rnd_value(0, buffers.len() - 1);
    buffers[index].clone()
}

pub fn play_sound_menu_move(core: &mut Core) {
    play_sound(core, core.sound_menu_move.clone());
}

pub fn play_sound_menu_select(core: &mut Core) {
    play_sound(core, core.sound_menu_select.clone());
}

pub fn play_sound_absorb(core: &mut Core) {
    play_sound(core, core.sound_absorb.clone());
}

pub fn play_sound_animal_throw(core: &mut Core) {
    play_sound(core, get_random_buffer(&core.sound_animal_throw));
}

pub fn play_sound_explosion(core: &mut Core) {
    play_sound(core, get_random_buffer(&core.sound_explosion));
}

pub fn play_sound_ground_impact(core: &mut Core) {
    play_sound(core, get_random_buffer(&core.sound_ground_impact));
}

pub fn play_sound_hurt(core: &mut Core) {
    play_sound(core, core.sound_hurt.clone());
}

pub fn play_sound_moon_rock_shoot(core: &mut Core) {
    play_sound(core, core.sound_moon_rock_shoot.clone());
}

pub fn play_sound_pickup_clover(core: &mut Core) {
    play_sound(core, core.sound_pickup_clover.clone());
}

pub fn play_sound_pickup_heart(core: &mut Core) {
    play_sound(core, core.sound_pickup_heart.clone());
}
