use libretrogd::entities::*;
use libretrogd::graphics::*;
use libretrogd::math::*;
use libretrogd::utils::rnd_value;

use crate::components::*;
use crate::events::*;
use crate::game::*;
use crate::logic::*;
use crate::tilemap::{TILE_HEIGHT, TILE_WIDTH};

///////////////////////////////////////////////////////////////////////////////////////////////////

pub fn update_system_world_tick(context: &mut Core) {
    if let Some((_, world)) = context.entities.components_mut::<World>().single_mut() {
        world.person_spawn_time -= context.delta;
        if world.person_spawn_time <= 0.0 {
            world.person_spawn_time = rnd_value(PERSON_SPAWN_TIMER_MIN, PERSON_SPAWN_TIMER_MAX);
            context.event_publisher.queue(Event::SpawnRandomPerson);
        }

        world.animal_spawn_time -= context.delta;
        if world.animal_spawn_time <= 0.0 {
            world.animal_spawn_time = rnd_value(ANIMAL_SPAWN_TIMER_MIN, ANIMAL_SPAWN_TIMER_MAX);
            context.event_publisher.queue(Event::SpawnRandomAnimal);
        } else {
            let no_animals = context.entities.components::<Animal>().len() == 0;
            if no_animals && world.animal_spawn_time >= ANIMAL_SPAWN_TIMER_PITY {
                world.animal_spawn_time = ANIMAL_SPAWN_TIMER_PITY;
            }
        }

        if let Some(mut remaining_time) = world.remaining_time.as_mut() {
            *remaining_time -= context.delta;
            if *remaining_time <= 0.0 {
                *remaining_time = 0.0;
            }
        }

        if let Some(mut time_to_game_over) = world.time_to_game_over.as_mut() {
            *time_to_game_over -= context.delta;
            if *time_to_game_over <= 0.0 {
                *time_to_game_over = 0.0;
            }
        }
    }
}

pub fn update_system_moon_tick(context: &mut Core) {
    let player_exists = get_player_entity_id(&context.entities).is_some();

    if let Some((moon_entity_id, mut moon)) = context.entities.components_mut::<Moon>().single_mut() {
        if moon.update(context.delta) {
            let mut color_offsets = context.entities.components_mut::<ColorOffset>();
            if let Some(color_offset) = color_offsets.get_mut(moon_entity_id) {
                color_offset.0 = moon.anger_color_offset();
            }

            if moon.should_shoot() && player_exists {
                context.event_publisher.queue(Event::Shoot(*moon_entity_id));
            }
        }
    }
}

pub fn update_system_movement(context: &mut Core) {
    if let Some(mut positions) = context.entities.components_mut::<Position>() {
        let velocities = context.entities.components::<Velocity>();
        let forces = context.entities.components::<Forces>();
        let bounds = context.entities.components::<Bounds>();
        let ignores_collision = context.entities.components::<IgnoresCollision>();

        for (entity, position) in positions.iter_mut() {
            if ignores_collision.contains_key(entity) {
                if let Some(velocity) = velocities.get(entity) {
                    position.0 += velocity.0 * context.delta;
                }
                if let Some(force) = forces.get(entity) {
                    position.0 += force.current_force();
                }
            } else {
                let velocity = velocities.get(entity);
                let force = forces.get(entity);
                if let Some(bound) = bounds.get(entity) {
                    move_entity_with_collision(
                        position,
                        bound,
                        velocity,
                        force,
                        &context.tilemap,
                        context.delta
                    );
                }
            }
        }
    }
}

pub fn update_system_flung(context: &mut Core) {
    if let Some(mut flungs) = context.entities.components_mut::<Flung>() {
        let velocities = context.entities.components::<Velocity>();
        let mut transforms = context.entities.components_mut::<Transform>();

        for (entity, flung) in flungs.iter_mut() {
            if let Some(velocity) = velocities.get(entity) {
                if flung.update(&velocity.0, context.delta) {
                    context.event_publisher.queue(Event::FinishedFlung(*entity));
                }
            }
            if let Some(mut transform) = transforms.get_mut(entity) {
                let travelled = flung.travelled();
                let scale = (travelled * RADIANS_180).sin() + 1.0;
                let rotation = travelled * 4.0 * RADIANS_360;
                transform.scale_x = scale;
                transform.scale_y = scale;
                transform.angle = rotation;
            }
        }
    }
}

pub fn update_system_friction(context: &mut Core) {
    if let Some(mut velocities) = context.entities.components_mut::<Velocity>() {
        let ignores_friction = context.entities.components::<IgnoresFriction>();

        for (entity, velocity) in velocities.iter_mut() {
            if !ignores_friction.contains_key(entity) {
                velocity.0 *= FRICTION;
                if velocity.0.almost_zero(0.001) {
                    velocity.0 = Vector2::ZERO;
                }
            }
        }
    }
}

pub fn update_system_force_decay(context: &mut Core) {
    if let Some(mut forces) = context.entities.components_mut::<Forces>() {
        for (_, force) in forces.iter_mut() {
            force.decay();
        }
    }
}

pub fn update_system_lifetime(context: &mut Core) {
    if let Some(mut lifetimes) = context.entities.components_mut::<LifeTime>() {
        for (entity, lifetime) in lifetimes.iter_mut() {
            if lifetime.update(context.delta) {
                context.event_publisher.queue(Event::Remove(*entity));
            }
        }
    }
}

pub fn update_system_animation(context: &mut Core) {
    if let Some(mut animations) = context.entities.components_mut::<AnimationInstance>() {
        let kill_when_animation_finishes = context.entities.components::<KillWhenAnimationFinishes>();

        for (entity, animation) in animations.iter_mut() {
            if animation.complete {
                continue;
            }

            animation.frame_timer += context.delta;

            let delay = if let Some(delay_override) = animation.delay_override {
                delay_override
            } else {
                animation.def.delay
            };

            if animation.frame_timer >= delay {
                // move to the next frame in the current sequence
                animation.frame_timer = 0.0;
                if animation.frame_index == (animation.def.frames.len() - 1) {
                    // we're at the last frame in the current sequence
                    if !animation.def.loops {
                        animation.complete = true;

                        context.event_publisher.queue(Event::AnimationFinished(*entity));
                        if kill_when_animation_finishes.contains_key(entity) {
                            context.event_publisher.queue(Event::Remove(*entity));
                        }
                    } else {
                        animation.frame_index = 0;
                    }
                } else {
                    animation.frame_index += 1;
                }
            }
        }
    }
}

pub fn update_system_timed_flicker(context: &mut Core) {
    if let Some(mut timed_flickers) = context.entities.components_mut::<TimedFlicker>() {
        for (_, flicker) in timed_flickers.iter_mut() {
            flicker.update(context.delta);
        }
        timed_flickers.retain(|_, flicker| flicker.timer > 0.0);
    }
}

pub fn update_system_transforms(context: &mut Core) {
    if let Some(mut transforms) = context.entities.components_mut::<Transform>() {
        let constant_rotations = context.entities.components::<ConstantRotation>();

        for (entity, transform) in transforms.iter_mut() {
            if let Some(constant_rotation) = constant_rotations.get(entity) {
                transform.angle = transform.angle.wrapping_radians_add(constant_rotation.0);
            }
        }
    }
}

pub fn update_system_walking_time(context: &mut Core) {
    if let Some(mut walking_times) = context.entities.components_mut::<WalkingTime>() {
        let activities = context.entities.components::<Activity>();

        for (entity, walking_time) in walking_times.iter_mut() {
            if let Some(activity) = activities.get(entity) {
                // dead entities can't walk!
                if activity.0 == EntityActivity::Dead {
                    continue;
                }
            }
            if walking_time.0 > 0.0 {
                walking_time.0 -= context.delta;
                context.event_publisher.queue(Event::MoveForward(*entity));
            }
        }

        // remove walking time components whose timers have elapsed
        walking_times.retain(|_, comp| comp.0 > 0.0);
    }
}

pub fn update_system_randomly_walk_around(context: &mut Core) {
    if let Some(mut randomly_walk_arounds) = context.entities.components_mut::<RandomlyWalksAround>() {
        let activities = context.entities.components::<Activity>();
        let mut walking_times = context.entities.components_mut::<WalkingTime>().unwrap();

        for (entity, randomly_walk_around) in randomly_walk_arounds.iter_mut() {
            if let Some(activity) = activities.get(entity) {
                if activity.0 == EntityActivity::Idle {
                    if randomly_walk_around.cooldown_timer > 0.0 {
                        randomly_walk_around.cooldown_timer -= context.delta;
                        if randomly_walk_around.cooldown_timer < 0.0 {
                            randomly_walk_around.cooldown_timer = 0.0;
                        }
                    } else if randomly_walk_around.should_start_walking() {
                        randomly_walk_around.cooldown_timer = rnd_value(randomly_walk_around.min_cooldown, randomly_walk_around.max_cooldown);

                        let direction = get_random_cardinal_direction_angle();
                        let walk_time = rnd_value(randomly_walk_around.min_walk_time, randomly_walk_around.max_walk_time);

                        walking_times.insert(*entity, WalkingTime(walk_time));
                        context.event_publisher.queue(Event::TurnAndMove(*entity, direction));
                    }
                }
            }
        }
    }
}

pub fn update_system_walk_constantly(context: &mut Core) {
    if let Some(walk_constantly_in_directions) = context.entities.components::<WalkConstantlyInDirection>() {
        for (entity, walk_constantly_in_direction) in walk_constantly_in_directions.iter() {
            context.event_publisher.queue(Event::TurnAndMove(*entity, walk_constantly_in_direction.0));
        }
    }
}

pub fn update_system_despawn_people_in_despawn_zone(context: &mut Core) {
    if let Some(peoples) = context.entities.components::<Person>() {
        let positions = context.entities.components::<Position>();
        for (entity, _) in peoples.iter() {
            let position = positions.get(entity).unwrap();
            // get x/y tile coordinates from the entity's center (not their top-left)
            let x_tile = (position.0.x as i32 + (TILE_WIDTH as i32 / 2)) / TILE_WIDTH as i32;
            let y_tile = (position.0.y as i32 + (TILE_HEIGHT as i32 / 2)) / TILE_HEIGHT as i32;
            if context.tilemap.is_in_people_despawn_zone(x_tile, y_tile) {
                context.event_publisher.queue(Event::Remove(*entity));
            }
        }
    }
}

pub fn update_system_check_target_positions(context: &mut Core) {
    if let Some(target_positions) = context.entities.components::<TargetPosition>() {
        let positions = context.entities.components::<Position>();
        let bounds = context.entities.components::<Bounds>();

        for (entity, target_position) in target_positions.iter() {
            let position = positions.get(entity).unwrap();
            let bounds = bounds.get(entity).unwrap();

            let rect = Rect::new(position.0.x as i32, position.0.y as i32, bounds.width, bounds.height);
            if rect.contains_point(target_position.0.x as i32, target_position.0.y as i32) {
                context.event_publisher.queue(Event::TargetReached(*entity));
            }
        }
    }
}

pub fn update_system_current_entity_activity(context: &mut Core) {
    if let Some(activities) = context.entities.components::<Activity>() {
        let velocities = context.entities.components::<Velocity>();

        for (entity, activity) in activities.iter() {
            // try to detect current entity activity based on it's own movement speed
            // (intentionally NOT checking force velocity!)
            if let Some(velocity) = velocities.get(entity) {
                match activity.0 {
                    EntityActivity::Idle => {
                        if velocity.0.length_squared() > 0.0 {
                            context.event_publisher.queue(Event::SetActivity {
                                entity: *entity,
                                activity: EntityActivity::Walking,
                            });
                        }
                    }
                    EntityActivity::Walking => {
                        if velocity.0.almost_zero(0.001) {
                            context.event_publisher.queue(Event::SetActivity {
                                entity: *entity,
                                activity: EntityActivity::Idle,
                            });
                        }
                    }
                    _ => {}
                }
            }
        }
    }
}

pub fn update_system_channeling(context: &mut Core) {
    let positions = context.entities.components::<Position>();
    let bounds = context.entities.components::<Bounds>();
    let mut channelers = context.entities.components_mut::<Channeler>().unwrap();
    let mut channelables = context.entities.components_mut::<Channelable>().unwrap();

    // TODO: this is slow

    for (channeling_entity, channeler) in channelers.iter_mut() {
        if channeler.update(context.delta) {
            context.event_publisher.queue(Event::StopChanneling(*channeling_entity));
        }

        if channeler.time > 0.0 {
            let channel_kind = channeler.kind.unwrap();
            let channeler_position = positions.get(channeling_entity).unwrap();
            let channeler_bounds = bounds.get(channeling_entity).unwrap();
            let channel_source_circle = Circle::new(
                channeler_position.0.x as i32 + (channeler_bounds.width / 2) as i32,
                channeler_position.0.y as i32 + (channeler_bounds.height / 2) as i32,
                channel_kind.distance() as u32,
            );
            let source_circle = circle_from_position_and_bounds(channeler_position, channeler_bounds);

            for (channelable_entity, channelable) in channelables.iter_mut().filter(|(e, c)| c.allowed_kind == channel_kind) {
                // shouldn't happen, but just in case
                if *channeling_entity == *channelable_entity {
                    continue;
                }

                let channelable_position = positions.get(channelable_entity).unwrap();
                let channelable_bounds = bounds.get(channelable_entity).unwrap();
                let channelable_circle = circle_from_position_and_bounds(channelable_position, channelable_bounds);

                // is the channeled entity now touching the channeler entity itself?
                if channelable.is_channeled() && source_circle.overlaps(&channelable_circle) {
                    context.event_publisher.queue(Event::ChannelledAtChanneler {
                        channeler: *channeling_entity,
                        channeled: *channelable_entity,
                        kind: channel_kind,
                    });
                }

                // is this potential entity within the bounds of the channeler's effect radius?
                if channel_source_circle.overlaps(&channelable_circle) {
                    if !channelable.is_channeled() {
                        context.event_publisher.queue(Event::Channelled {
                            channeler: *channeling_entity,
                            victim: *channelable_entity,
                        });
                    }
                    channelable.refresh(channeler_position.0);
                }
            }
        }
    }
}

pub fn update_system_move_channeled_entities(context: &mut Core) {
    if let Some(mut channelled_entities) = context.entities.components_mut::<Channelable>() {
        let mut forces = context.entities.components_mut::<Forces>();
        let positions = context.entities.components::<Position>();

        for (entity, channelable) in channelled_entities.iter_mut() {
            if channelable.update(context.delta) {
                context.event_publisher.queue(Event::StopChanneled(*entity));
            }

            if channelable.time > 0.0 {
                let position = positions.get(entity).unwrap();
                let force = forces.get_mut(entity).unwrap();
                let force_direction = (channelable.target - position.0).normalize();
                let v = force_direction * channelable.allowed_kind.force_strength();
                let dissipation = CHANNELLED_FORCE_DISSIPATION;
                force.add(v, dissipation);
            }
        }
    }
}

pub fn update_system_projectile_hits(context: &mut Core) {
    if let Some(projectiles) = context.entities.components::<Projectile>() {
        if let Some(shootables) = context.entities.components::<Shootable>() {
            let positions = context.entities.components::<Position>();
            let bounds = context.entities.components::<Bounds>();

            // TODO: this is slow

            for (projectile_entity, _) in projectiles.iter() {
                let projectile_position = positions.get(projectile_entity).unwrap();
                let projectile_bounds = bounds.get(projectile_entity).unwrap();
                let projectile_circle = circle_from_position_and_bounds(projectile_position, projectile_bounds);

                let mut hit_something = false;
                for (shootable_entity, _) in shootables.iter() {
                    let shootable_position = positions.get(shootable_entity).unwrap();
                    let shootable_bounds = bounds.get(shootable_entity).unwrap();
                    let shootable_circle = circle_from_position_and_bounds(shootable_position, shootable_bounds);

                    if projectile_circle.overlaps(&shootable_circle) {
                        hit_something = true;
                        context.event_publisher.queue(Event::Shot {
                            victim: *shootable_entity,
                            projectile: *projectile_entity,
                        });
                    }
                }

                if hit_something {
                    context.event_publisher.queue(Event::Hurt {
                        victim: *projectile_entity,
                        source: None,
                        amount: 1
                    });
                }
            }
        }
    }
}

pub fn update_system_pushing(context: &mut Core) {
    if let Some(pushers) = context.entities.components::<Pusher>() {
        let positions = context.entities.components::<Position>();
        let bounds = context.entities.components::<Bounds>();
        let mut forces = context.entities.components_mut::<Forces>();
        let pushable = context.entities.components::<Pushable>().unwrap();

        // TODO: this is slow

        for (pusher_entity, pusher) in pushers.iter() {
            let pusher_position = positions.get(pusher_entity).unwrap();
            let pusher_bounds = bounds.get(pusher_entity).unwrap();
            let pusher_circle = circle_from_position_and_bounds(pusher_position, pusher_bounds);

            for (pushable_entity, pushable) in pushable.iter() {
                // don't push ourself ...
                if *pushable_entity == *pusher_entity {
                    continue;
                }

                let pushable_position = positions.get(pushable_entity).unwrap();
                let pushable_bounds = bounds.get(pushable_entity).unwrap();
                let pushable_circle = circle_from_position_and_bounds(pushable_position, pushable_bounds);

                if pusher_circle.overlaps(&pushable_circle) {
                    let push_direction = (pushable_position.0 - pusher_position.0).normalize();

                    let pushable_force = forces.get_mut(pushable_entity).unwrap();
                    pushable_force.add(push_direction * pusher.strength, pusher.push_force_dissipation);
                }
            }
        }
    }
}

pub fn update_system_flip_sprite_horizontally_based_on_facing_direction(context: &mut Core) {
    if let Some(flips) = context.entities.components::<FlipHorizontallyBasedOnDirection>() {
        let mut sprites = context.entities.components_mut::<Sprite>();
        let mut facing_angles = context.entities.components::<FacingAngle>();

        for (entity, _) in flips.iter() {
            if let Some(sprite) = sprites.get_mut(entity) {
                if let Some(facing_angle) = facing_angles.get(entity) {
                    if facing_angle.0 > RADIANS_90 && facing_angle.0 < RADIANS_270 {
                        sprite.flip_horizontally = true
                    } else {
                        sprite.flip_horizontally = false
                    }
                }
            }
        }
    }
}

pub fn update_system_adjust_render_properties_from_animation(context: &mut Core) {
    if let Some(animations) = context.entities.components::<AnimationInstance>() {
        let mut sprites = context.entities.components_mut::<Sprite>();
        let mut pixels = context.entities.components_mut::<Pixel>();

        for (entity, animation) in animations.iter() {
            let value = animation.def.frames[animation.frame_index];
            if let Some(sprite) = sprites.get_mut(entity) {
                match animation.def.method {
                    AnimationMethod::SpriteIndex => sprite.index = value,
                    AnimationMethod::ColorOffset => sprite.color = value as u8
                }
            } else if let Some(pixel) = pixels.get_mut(entity) {
                match animation.def.method {
                    AnimationMethod::ColorOffset => pixel.0 = value as u8,
                    _ => {},
                }
            }
        }
    }
}

pub fn update_system_pickups(context: &mut Core) {
    if let Some(mut pickupables) = context.entities.components_mut::<Pickupable>() {
        let pickupers = context.entities.components::<Pickuper>().unwrap();
        let positions = context.entities.components::<Position>();
        let bounds = context.entities.components::<Bounds>();

        // don't really think this pre_timer thing is necessary anymore ... ?
        for (_, pickupable) in pickupables.iter_mut() {
            if pickupable.pre_timer > 0.0 {
                pickupable.pre_timer -= context.delta;
            }
        }

        // TODO: this is slow

        for (pickuper_entity, _) in pickupers.iter() {
            let pickuper_position = positions.get(pickuper_entity).unwrap();
            let pickuper_bounds = bounds.get(pickuper_entity).unwrap();
            let pickuper_circle = circle_from_position_and_bounds(pickuper_position, pickuper_bounds);

            for (pickupable_entity, pickupable) in pickupables.iter() {
                if pickupable.pre_timer <= 0.0 {
                    let pickupable_position = positions.get(pickupable_entity).unwrap();
                    let pickupable_bounds = bounds.get(pickupable_entity).unwrap();
                    let pickupable_circle = circle_from_position_and_bounds(pickupable_position, pickupable_bounds);

                    if pickupable_circle.overlaps(&pickuper_circle) {
                        context.event_publisher.queue(Event::Pickup {
                            picked_up_by: *pickuper_entity,
                            picked_up: *pickupable_entity
                        });
                    }
                }
            }
        }
    }
}

pub fn update_system_particle_emitter(context: &mut Core) {
    if let Some(mut emitters) = context.entities.components_mut::<ParticleEmitter>() {
        for (entity, emitter) in emitters.iter_mut() {
            if emitter.update(context.delta) {
                emitter.reset();
                context.event_publisher.queue(Event::EmitParticle {
                    source: *entity,
                    kind: emitter.kind,
                    count: emitter.emit_count(),
                });
            }
        }
    }
}

pub fn update_system_life_and_death(context: &mut Core) {
    if let Some(lifes) = context.entities.components::<Life>() {
        for (entity, life) in lifes.iter() {
            if life.0 <= 0 {
                context.event_publisher.queue(Event::Killed(*entity));
            }
        }
    }
}

///////////////////////////////////////////////////////////////////////////////////////////////////

pub fn render_system_world_base(context: &mut Core) {
    if let Some((entity, world)) = context.entities.components::<World>().single() {
        let cameras = context.entities.components::<Camera>();
        let camera = cameras.get(entity).unwrap();
        context.tilemap.draw_base(&mut context.system.video, &context.tiles, camera.x, camera.y);
    }
}

pub fn render_system_world_overlay(context: &mut Core) {
    if let Some((entity, world)) = context.entities.components::<World>().single() {
        let cameras = context.entities.components::<Camera>();
        let camera = cameras.get(entity).unwrap();
        context.tilemap.draw_overlay(&mut context.system.video, &context.tiles, camera.x, camera.y);
    }
}

fn get_final_entity_color(
    entity: &EntityId,
    initial_color: u8,
    color_offsets: &RefComponents<ColorOffset>,
    lifetimes: &RefComponents<LifeTime>,
    color_by_lifetimes: &RefComponents<ColorByLifeTime>,
) -> u8 {
    let mut color = initial_color;
    if let Some(color_offset) = color_offsets.get(entity) {
        color = color.wrapping_add(color_offset.0);
    } else if let Some(color_by_lifetime) = color_by_lifetimes.get(entity) {
        let lifetime = lifetimes.get(entity).unwrap();
        color = color_by_lifetime.color(lifetime.life_percent());
    }
    color
}

fn render_sprite_entity(
    bitmap: &mut Bitmap,
    entity: &EntityId,
    sprite: &Sprite,
    positions: &RefComponents<Position>,
    color_offsets: &RefComponents<ColorOffset>,
    timed_flickers: &RefComponents<TimedFlicker>,
    transforms: &RefComponents<Transform>,
    lifetimes: &RefComponents<LifeTime>,
    color_by_lifetimes: &RefComponents<ColorByLifeTime>,
    camera: &Camera,
) {
    if let Some(position) = positions.get(entity) {
        let color = get_final_entity_color(entity, sprite.color, color_offsets, lifetimes, color_by_lifetimes);
        let flipped = sprite.flip_vertically || sprite.flip_horizontally;

        let blit_method;

        if let Some(flicker) = timed_flickers.get(entity) {
            if !flicker.flick {
                match flicker.method {
                    FlickerMethod::OnOff => {
                        // nothing to draw, currently not visible due to flicker state
                        return;
                    },
                    FlickerMethod::Color(draw_color) => {
                        if flipped {
                            blit_method = BlitMethod::TransparentFlippedSingle {
                                transparent_color: 0,
                                horizontal_flip: sprite.flip_horizontally,
                                vertical_flip: sprite.flip_vertically,
                                draw_color,
                            };
                        } else {
                            blit_method = BlitMethod::TransparentSingle {
                                transparent_color: 0,
                                draw_color,
                            };
                        }
                    }
                }
            } else {
                if flipped {
                    blit_method = BlitMethod::TransparentFlippedOffset {
                        transparent_color: 0,
                        horizontal_flip: sprite.flip_horizontally,
                        vertical_flip: sprite.flip_vertically,
                        offset: color,
                    };
                } else {
                    blit_method = BlitMethod::TransparentOffset {
                        transparent_color: 0,
                        offset: color,
                    };
                }
            }
        } else if let Some(transform) = transforms.get(entity) {
            blit_method = BlitMethod::RotoZoomTransparentOffset {
                transparent_color: 0,
                offset: color,
                angle: transform.angle,
                scale_x: transform.scale_x,
                scale_y: transform.scale_y,
            };
        } else {
            if flipped {
                blit_method = BlitMethod::TransparentFlippedOffset {
                    transparent_color: 0,
                    horizontal_flip: sprite.flip_horizontally,
                    vertical_flip: sprite.flip_vertically,
                    offset: color,
                };
            } else {
                blit_method = BlitMethod::TransparentOffset {
                    transparent_color: 0,
                    offset: color,
                };
            }
        }

        bitmap.blit_region(
            blit_method,
            sprite.atlas.bitmap(),
            &sprite.atlas[sprite.index],
            position.0.x as i32 - camera.x,
            position.0.y as i32 - camera.y
        );
    }
}

pub fn render_system_render_sprites(context: &mut Core) {
    if let Some(sprites) = context.entities.components::<Sprite>() {
        if let Some((_, camera)) = context.entities.components::<Camera>().single() {
            let positions = context.entities.components::<Position>().unwrap();
            let color_offsets = context.entities.components::<ColorOffset>().unwrap();
            let timed_flickers = context.entities.components::<TimedFlicker>().unwrap();
            let transforms = context.entities.components::<Transform>().unwrap();
            let above_overlays = context.entities.components::<RendersAboveTilemapOverlay>().unwrap();
            let lifetimes = context.entities.components::<LifeTime>().unwrap();
            let color_by_lifetimes = context.entities.components::<ColorByLifeTime>().unwrap();

            for (entity, sprite) in sprites.iter() {
                if !above_overlays.contains_key(entity) {
                    render_sprite_entity(
                        &mut context.system.video,
                        entity,
                        sprite,
                        &positions,
                        &color_offsets,
                        &timed_flickers,
                        &transforms,
                        &lifetimes,
                        &color_by_lifetimes,
                        camera
                    );
                }
            }
        }
    }
}

pub fn render_system_render_sprites_top(context: &mut Core) {
    if let Some(sprites) = context.entities.components::<Sprite>() {
        if let Some((_, camera)) = context.entities.components::<Camera>().single() {
            let positions = context.entities.components::<Position>().unwrap();
            let color_offsets = context.entities.components::<ColorOffset>().unwrap();
            let timed_flickers = context.entities.components::<TimedFlicker>().unwrap();
            let transforms = context.entities.components::<Transform>().unwrap();
            let above_overlays = context.entities.components::<RendersAboveTilemapOverlay>().unwrap();
            let lifetimes = context.entities.components::<LifeTime>().unwrap();
            let color_by_lifetimes = context.entities.components::<ColorByLifeTime>().unwrap();

            for (entity, sprite) in sprites.iter() {
                if above_overlays.contains_key(entity) {
                    render_sprite_entity(
                        &mut context.system.video,
                        entity,
                        sprite,
                        &positions,
                        &color_offsets,
                        &timed_flickers,
                        &transforms,
                        &lifetimes,
                        &color_by_lifetimes,
                        camera
                    );
                }
            }
        }
    }
}

fn render_pixel_entity(
    bitmap: &mut Bitmap,
    entity: &EntityId,
    pixel: &Pixel,
    positions: &RefComponents<Position>,
    color_offsets: &RefComponents<ColorOffset>,
    lifetimes: &RefComponents<LifeTime>,
    color_by_lifetimes: &RefComponents<ColorByLifeTime>,
    camera: &Camera,
) {
    if let Some(position) = positions.get(entity) {
        let color = get_final_entity_color(entity, pixel.0, color_offsets, lifetimes, color_by_lifetimes);

        bitmap.set_pixel(
            position.0.x as i32 - camera.x,
            position.0.y as i32 - camera.y,
            color
        );
    }
}

pub fn render_system_render_pixels(context: &mut Core) {
    if let Some(pixels) = context.entities.components::<Pixel>() {
        if let Some((_, camera)) = context.entities.components::<Camera>().single() {
            let positions = context.entities.components::<Position>().unwrap();
            let color_offsets = context.entities.components::<ColorOffset>().unwrap();
            let above_overlays = context.entities.components::<RendersAboveTilemapOverlay>().unwrap();
            let lifetimes = context.entities.components::<LifeTime>().unwrap();
            let color_by_lifetimes = context.entities.components::<ColorByLifeTime>().unwrap();

            for (entity, pixel) in pixels.iter() {
                if !above_overlays.contains_key(entity) {
                    render_pixel_entity(
                        &mut context.system.video,
                        entity,
                        pixel,
                        &positions,
                        &color_offsets,
                        &lifetimes,
                        &color_by_lifetimes,
                        camera
                    );
                }
            }
        }
    }
}

pub fn render_system_render_pixels_top(context: &mut Core) {
    if let Some(pixels) = context.entities.components::<Pixel>() {
        if let Some((_, camera)) = context.entities.components::<Camera>().single() {
            let positions = context.entities.components::<Position>().unwrap();
            let color_offsets = context.entities.components::<ColorOffset>().unwrap();
            let above_overlays = context.entities.components::<RendersAboveTilemapOverlay>().unwrap();
            let lifetimes = context.entities.components::<LifeTime>().unwrap();
            let color_by_lifetimes = context.entities.components::<ColorByLifeTime>().unwrap();

            for (entity, pixel) in pixels.iter() {
                if above_overlays.contains_key(entity) {
                    render_pixel_entity(
                        &mut context.system.video,
                        entity,
                        pixel,
                        &positions,
                        &color_offsets,
                        &lifetimes,
                        &color_by_lifetimes,
                        camera
                    );
                }
            }
        }
    }
}

pub fn render_system_effects(context: &mut Core) {
    if let Some(channelers) = context.entities.components::<Channeler>() {
        if let Some((_, camera)) = context.entities.components::<Camera>().single() {
            let positions = context.entities.components::<Position>().unwrap();
            let bounds = context.entities.components::<Bounds>().unwrap();

            for (entity, channeler) in channelers.iter() {
                if channeler.is_channeling() {
                    let color = channeler.kind.unwrap().color();
                    let position = positions.get(entity).unwrap();
                    let bound = bounds.get(entity).unwrap();

                    let radius = lerp(0.0, channeler.kind.unwrap().distance(), channeler.cycle / CHANNEL_CYCLE_TIME);
                    let circle = Circle::new(
                        position.0.x as i32 + (bound.width / 2) as i32 - camera.x,
                        position.0.y as i32 + (bound.height / 2) as i32 - camera.y,
                        radius as u32
                    );
                    context.system.video.circle(circle.x, circle.y, circle.radius, color);
                }
            }
        }
    }
}