use std::fs::File;
use std::io::BufReader;
use std::path::Path;

use anyhow::{Context, Result};
use byteorder::ReadBytesExt;
use libretrogd::graphics::*;
use libretrogd::math::*;
use libretrogd::utils::rnd_value;
use serde::Deserialize;

pub const TILE_WIDTH: u32 = 8;
pub const TILE_HEIGHT: u32 = 8;

pub const TILE_FLAG_COLLISION: i32 = 1;
pub const TILE_FLAG_NOTHING: i32 = 0;
pub const TILE_FLAG_PEOPLE_SPAWN: i32 = -1;
pub const TILE_FLAG_PEOPLE_DESPAWN: i32 = -2;
pub const TILE_FLAG_ANIMAL_SPAWN: i32 = -3;

#[derive(Deserialize, Clone)]
pub struct TileMap {
    width: u32,
    height: u32,
    layers: Vec<Box<[i32]>>,

    #[serde(skip)]
    people_spawn_positions: Vec<(i32, i32)>,
    #[serde(skip)]
    people_despawn_positions: Vec<(i32, i32)>,
    #[serde(skip)]
    animal_spawn_positions: Vec<(i32, i32)>,
}

impl TileMap {
    pub fn load_file(path: &Path) -> Result<Self> {
        let f = File::open(path).context(format!("Opening json tilemap: {:?}", path))?;
        let mut reader = BufReader::new(f);
        Self::load_bytes(&mut reader)
    }

    pub fn load_bytes<T: ReadBytesExt>(reader: &mut T) -> Result<Self> {
        let mut tilemap: TileMap = serde_json::from_reader(reader).context("Loading json tilemap bytes")?;

        for y in 0..tilemap.height as i32 {
            for x in 0..tilemap.width as i32 {
                let offset = tilemap.index_to(x, y).unwrap();
                match tilemap.flags()[offset] {
                    TILE_FLAG_PEOPLE_SPAWN => tilemap.people_spawn_positions.push((x, y)),
                    TILE_FLAG_PEOPLE_DESPAWN => tilemap.people_despawn_positions.push((x, y)),
                    TILE_FLAG_ANIMAL_SPAWN => tilemap.animal_spawn_positions.push((x, y)),
                    _ => {}
                }
            }
        }

        Ok(tilemap)
    }

    #[inline]
    pub fn width(&self) -> u32 {
        self.width
    }

    #[inline]
    pub fn height(&self) -> u32 {
        self.height
    }

    #[inline]
    pub fn lower_tiles(&self) -> &Box<[i32]> {
        &self.layers[0]
    }

    #[inline]
    pub fn lower_colors(&self) -> &Box<[i32]> {
        &self.layers[1]
    }

    #[inline]
    pub fn upper_tiles(&self) -> &Box<[i32]> {
        &self.layers[2]
    }

    #[inline]
    pub fn upper_colors(&self) -> &Box<[i32]> {
        &self.layers[3]
    }

    #[inline]
    pub fn overlay_tiles(&self) -> &Box<[i32]> {
        &self.layers[4]
    }

    #[inline]
    pub fn overlay_colors(&self) -> &Box<[i32]> {
        &self.layers[5]
    }

    #[inline]
    pub fn flags(&self) -> &Box<[i32]> {
        &self.layers[6]
    }

    pub fn index_to(&self, x: i32, y: i32) -> Option<usize> {
        if x >= 0 && y >= 0 && x < self.width as i32 && y < self.height as i32 {
            Some(((y * self.width as i32) + x) as usize)
        } else {
            None
        }
    }

    #[inline]
    fn get_draw_values(&self, camera_x: i32, camera_y: i32) -> (i32, i32, i32, i32) {
        let xt = camera_x / TILE_WIDTH as i32;
        let yt = camera_y / TILE_HEIGHT as i32;
        let xp = camera_x % TILE_WIDTH as i32;
        let yp = camera_y % TILE_HEIGHT as i32;
        (xt, yt, xp, yp)
    }

    #[inline]
    fn get_tile_draw_position(
        &self,
        x_screen_tile: i32,
        y_screen_tile: i32,
        x_pixel_offset: i32,
        y_pixel_offset: i32
    ) -> (i32, i32) {
        let xd = (x_screen_tile * TILE_WIDTH as i32) - x_pixel_offset;
        let yd = (y_screen_tile * TILE_HEIGHT as i32) - y_pixel_offset;
        (xd, yd)
    }

    pub fn draw_base(&self, dest: &mut Bitmap, tiles: &BitmapAtlas, camera_x: i32, camera_y: i32) {
        let (xt, yt, xp, yp) = self.get_draw_values(camera_x, camera_y);

        for y in 0..=15 {
            for x in 0..=20 {
                if let Some(index) = self.index_to(x + xt, y + yt) {
                    let (xd, yd) = self.get_tile_draw_position(x, y, xp, yp);

                    let lower_tile = self.lower_tiles()[index];
                    let lower_color = self.lower_colors()[index] as u8;
                    if lower_tile >= 0 {
                        dest.blit_region(
                            BlitMethod::SolidOffset(lower_color),
                            tiles.bitmap(),
                            &tiles[lower_tile as usize],
                            xd,
                            yd,
                        )
                    }

                    let upper_tile = self.upper_tiles()[index];
                    let upper_color = self.upper_colors()[index] as u8;
                    if upper_tile >= 0 {
                        dest.blit_region(
                            BlitMethod::TransparentOffset {
                                transparent_color: 0,
                                offset: upper_color,
                            },
                            tiles.bitmap(),
                            &tiles[upper_tile as usize],
                            xd,
                            yd,
                        )
                    }
                }
            }
        }
    }

    pub fn draw_overlay(&self, dest: &mut Bitmap, tiles: &BitmapAtlas, camera_x: i32, camera_y: i32) {
        let (xt, yt, xp, yp) = self.get_draw_values(camera_x, camera_y);

        for y in 0..=15 {
            for x in 0..=20 {
                if let Some(index) = self.index_to(x + xt, y + yt) {
                    let (xd, yd) = self.get_tile_draw_position(x, y, xp, yp);

                    let overlay_tile = self.overlay_tiles()[index];
                    let overlay_color = self.overlay_colors()[index] as u8;
                    if overlay_tile >= 0 {
                        dest.blit_region(
                            BlitMethod::TransparentOffset {
                                transparent_color: 0,
                                offset: overlay_color,
                            },
                            tiles.bitmap(),
                            &tiles[overlay_tile as usize],
                            xd,
                            yd
                        )
                    }
                }
            }
        }
    }

    pub fn is_colliding(&self, rect: &Rect) -> bool {
        let x1 = rect.x / TILE_WIDTH as i32;
        let y1 = rect.y / TILE_HEIGHT as i32;
        let x2 = rect.right() / TILE_WIDTH as i32;
        let y2 = rect.bottom() / TILE_HEIGHT as i32;

        for y in y1..=y2 {
            for x in x1..=x2 {
                match self.index_to(x, y) {
                    Some(index) => {
                        if self.flags()[index] > 0 {
                            return true;
                        }
                    },
                    None => return true
                }
            }
        }
        false
    }

    pub fn get_random_people_spawn_coordinate(&self) -> Option<(i32, i32)> {
        if self.people_spawn_positions.len() > 0 {
            let index = rnd_value(0, self.people_spawn_positions.len() - 1);
            Some(self.people_spawn_positions[index])
        } else {
            None
        }
    }

    pub fn get_random_animal_spawn_coordinate(&self) -> Option<(i32, i32)> {
        if self.animal_spawn_positions.len() > 0 {
            let index = rnd_value(0, self.animal_spawn_positions.len() - 1);
            Some(self.animal_spawn_positions[index])
        } else {
            None
        }
    }

    pub fn is_in_people_despawn_zone(&self, x_tile: i32, y_tile: i32) -> bool {
        for (x, y) in self.people_despawn_positions.iter() {
            if x_tile == *x && y_tile == *y {
                return true
            }
        }
        false
    }
}
